/*
 * CommunityDictUtils.java
 *
 * Created on September 2, 2007, 8:00 PM
 *
 */

package info.heinzelnisse.dictionary;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.List;
import java.util.Set;
import java.util.logging.Logger;

/**
 *
 * @author Heiko Klein
 */
public class CommunityDictUtils extends DictUtils {
    static CommunityDictUtils instance;
    
    protected CommunityDictUtils() {
        // singleton
    }

    public static synchronized CommunityDictUtils getInstance() {
        if (instance == null) {
            instance = new CommunityDictUtils();
        }
        return instance;
    }

    public String getTablePrefix() {
        return "Community";
    }

    /**
     * insert the transation
     * @param translation with lang == DE
     * @return translationId of this insert, or -1 on failure
     */
    private int insertTranslationIntoDb(CommunityTranslation translation) {
        int translationId = CommunityTranslation.NOT_IN_DATABASE;
        Connection conn = null;
        PreparedStatement statement = null;
        ResultSet rs = null;
            
        try {
            conn = DictUtils.dbConnection();

            String query = "INSERT INTO " + dbName(DictUtils.Table.TRANSLATION) +
                    " (word_DE, article_DE, other_DE,  word_NO, article_NO, other_NO, category, grade, forumUser, comment, flag, address, originalTranslationId) VALUES " +
                    " (      ?,          ?,        ?,        ?,          ?,        ?,        ?,     ?,         ?,       ?,    ?,       ?,                     ?)";
                  //" (      1,          2,        3,        4,          5,        6,        7,     8,         9,      10,   11,      12,                    13)";
            statement = conn.prepareStatement(query, PreparedStatement.RETURN_GENERATED_KEYS);
            statement.setString(1, translation.getWord());
            statement.setString(2, translation.getArticle());
            statement.setString(3, translation.getOther());
            statement.setString(4, translation.getT_word());
            statement.setString(5, translation.getT_article());
            statement.setString(6, translation.getT_other());
            statement.setString(7, translation.getCategory());
            statement.setString(8, translation.getGrade());
            statement.setInt(9, translation.getForumUser());
            statement.setString(10, translation.getComment());
            statement.setString(11, translation.getFlag().toString());
            statement.setString(12, translation.getAddress());
            statement.setInt(13, translation.getOriginalTranslationId());
            statement.execute();
            rs = statement.getGeneratedKeys();
            if (rs.next()) {
                translationId = rs.getInt(1);
            }
        } catch (SQLException se) {
            se.printStackTrace();
        } finally {
            try {rs.close();} catch (Exception e) {}
            try {statement.close();} catch (Exception e) {}
            try {conn.close();} catch (Exception e) {}
        }
      
        return translationId;
    }
    
    /**
     * @param translation translation with language DE, translationId ignored and set in the returned Translation
     * originalTranslationId should be set to NOT_IN_DATABASE if this is a creation rather than an update
     */
    public CommunityTranslation createUpdateTranslation(CommunityTranslation translation) {
        if (translation.getLang() != DictUtils.DictLang.DE) {
            Logger.getLogger(CommunityDictUtils.class.getName()).warning("require DE translation, nothing created");
            return null;
        }
        int translationId = insertTranslationIntoDb(translation);
        if (translationId == CommunityTranslation.NOT_IN_DATABASE) {
            return null;   
        }
        int originalTranslationId = translation.getOriginalTranslationId();
        if (originalTranslationId == CommunityTranslation.NOT_IN_DATABASE) {
            // create !! should maybe move to a function of its own
            originalTranslationId = translationId;
            Connection conn = null;
            PreparedStatement statement = null;
            ResultSet rs = null;
            try {
                conn = DictUtils.dbConnection();
                String query = "UPDATE " + dbName(DictUtils.Table.TRANSLATION) + " SET originalTranslationId = ? " + " WHERE translationId = ? ";
                statement = conn.prepareStatement(query);
                statement.setInt(1, originalTranslationId);
                statement.setInt(2, translationId); // originalId = Id on insert
                statement.execute();
            } catch (SQLException se) {
                se.printStackTrace();
            } finally {
                try {
                    rs.close();
                } catch (Exception e) {
                }
                try {
                    statement.close();
                } catch (Exception e) {
                }
                try {
                    conn.close();
                } catch (Exception e) {
                }
            }
        }
        CommunityTranslation retVal = new CommunityTranslation(translation.getLang(), translationId, translation.getWord(), translation.getArticle(), translation.getOther(), translation.getT_word(), translation.getT_article(), translation.getT_other(), translation.getCategory(), translation.getGrade(), translation.getForumUser(), translation.getComment(), translation.getFlag(), translation.getTimestamp(), translation.getAddress(), originalTranslationId);
        updateIndices(retVal);
        return retVal;
    }

    /**
     * update the indices for the translation
     * currently, only the word and T_word are added to the standard-index
     * others are not included
     * no stemmer, no phonetics
     */
    private void updateIndices(CommunityTranslation translation) {
        if (translation.getLang() != DictUtils.DictLang.DE) {
            Logger.getLogger(CommunityDictUtils.class.getName()).warning("require DE translation, nothing created");
            return;
        }
        String deCleanQuery = "DELETE FROM CommunityIndexWord_DE WHERE translationId in (SELECT translationId FROM CommunityTranslation WHERE originalTranslationId = ?)";
        String noCleanQuery = "DELETE FROM CommunityIndexWord_NO WHERE translationId in (SELECT translationId FROM CommunityTranslation WHERE originalTranslationId = ?)";
        
        String deQuery = "INSERT INTO CommunityIndexWord_DE (indexWord, translationId) VALUES (?, ?)";
        String noQuery = "INSERT INTO CommunityIndexWord_NO (indexWord, translationId) VALUES (?, ?)";
        Connection conn = null;
        PreparedStatement statement = null;
        try {
            conn = DictUtils.dbConnection();
            statement = conn.prepareStatement(deCleanQuery);
            statement.setInt(1, translation.getOriginalTranslationId());
            statement.execute();
            statement.close();
            statement = conn.prepareStatement(noCleanQuery);
            statement.setInt(1, translation.getOriginalTranslationId());
            statement.execute();
            statement.close();

            switch (translation.getFlag()) {
                case rejected: break;
                case deleted: // deleted should be searchable, but marked
                case enabled:
                case accepted:
                default:
                    {
                        statement = conn.prepareStatement(deQuery);
                        for (String s : Utils.multiplyUmlautWords(Utils.cleanWord(translation.getWord().toLowerCase(Utils.langLocale(DictUtils.DictLang.DE))))) {
                            statement.setString(1, s);
                            statement.setInt(2, translation.getId());
                            statement.execute();
                        }
                        statement.close();
                        statement = conn.prepareStatement(noQuery);
                        for (String s : Utils.multiplyUmlautWords(Utils.cleanWord(translation.getT_word().toLowerCase(Utils.langLocale(DictUtils.DictLang.NO))))) {
                            statement.setString(1, s);
                            statement.setInt(2, translation.getId());
                            statement.execute();
                        }
                    }
            }

        } catch (SQLException se) {
            se.printStackTrace();
        } finally {
            try {statement.close();} catch (Exception e) {}
            try {conn.close();} catch (Exception e) {}
        }
    }
    
    /**
     * search database for all items with exactly searchitem
     * use WordHelper
     *
     * @return: sorted List of Translation
     */
    public List<CommunityTranslation> dbSearch(String searchItem, DictLang lang) {
        Set<Integer> uniqueIdSet = dbSearchIdSet(searchItem, lang);
        List<CommunityTranslation> translations = new ArrayList<CommunityTranslation>(getTranslations(uniqueIdSet, lang));
        Collections.sort(translations);
        return translations;
    }
    
    /**
     * search database for all items with searchitem approximately
     *
     * @return: sorted List of Translation
     */
    public List<CommunityTranslation> dbSearchStart(String searchItem, DictLang lang) {
        Set<Integer> uniqueIdSet = dbSearchStartIdSet(searchItem, lang);
        List<CommunityTranslation> translations = new ArrayList<CommunityTranslation>(getTranslations(uniqueIdSet, lang));
        Collections.sort(translations);
        return translations;
    }
    
    public Collection<CommunityTranslation> getTranslations(Collection<Integer> ids, DictUtils.DictLang language) {
        String lang = DictUtils.dbLangFlag(language);
        String tlang = DictUtils.dbTransLangFlag(language);
        
        Collection<CommunityTranslation> output = new ArrayList<CommunityTranslation>();
        if (ids.size() != 0) {
            Connection conn = null;
            PreparedStatement statement = null;
            ResultSet rs = null;
            try {
                conn = DictUtils.dbConnection();
                StringBuffer questionMarks = new StringBuffer();
                for (int i = 0; i < ids.size(); i++) {
                    questionMarks.append("?,");
                }
                questionMarks.deleteCharAt(questionMarks.length()-1); // remove last ,
                String query = "SELECT translationId, CommunityTranslation.word_"+lang+" AS word, CommunityTranslation.article_"+lang+" AS article, other_"+lang+" AS other, CommunityTranslation.word_"+tlang+" AS t_word, CommunityTranslation.article_"+tlang+" AS t_article, other_"+tlang+" AS t_other, category, grade, forumUser, comment, flag, timestamp, address, originalTranslationId "+
                        "FROM CommunityTranslation "+
                        "WHERE translationId IN ("+questionMarks.toString()+")";
                statement = conn.prepareStatement(query);
                int i = 1;
                for (int id : ids) {
                    statement.setInt(i, id);
                    i++;
                }
                rs = statement.executeQuery();
                while (rs.next()) {
                    output.add(new CommunityTranslation(language, rs.getInt("translationId"), rs.getString("word"), rs.getString("article"), rs.getString("other"), rs.getString("t_word"), rs.getString("t_article"), rs.getString("t_other"), rs.getString("category"), rs.getString("grade"), rs.getInt("forumUser"), rs.getString("comment"), CommunityTranslation.Flag.valueOf(rs.getString("flag")), new Date(rs.getDate("timestamp").getTime()), rs.getString("address"), rs.getInt("originalTranslationId")));
                }
            } catch (SQLException se) {
                se.printStackTrace();
            } finally {
                try {rs.close();} catch (Exception e) {}
                try {statement.close();} catch (Exception e) {}
                try {conn.close();} catch (Exception e) {}
            }
        }
        return output;
    }
        
}