/*
 * DbTranslationStore.java
 *
 * Created on August 23, 2006, 8:40 AM
 *
 */

package info.heinzelnisse.dictionary;

import info.heinzelnisse.multiphone.AbstractMultiphone;
import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Locale;
import java.util.Set;

/**
 *
 * The DbTranslationStore Singleton makes sure there is only one
 * instance with a valid db-connection. Several instances without
 * db-connection might exist (although they will be useless)
 *
 * @author Heiko Klein
 */
public class DbTranslationStore implements TranslationStore {
    transient Connection conn = null;
    transient Statement sth = null;
    int translationCounter = 0;
    static final String translationQuery = "INSERT INTO Translation (translationId, word_DE, article_DE, other_DE,  word_NO, article_NO, other_NO, category, grade) VALUES ";
    List<String> translationValues = new ArrayList<String>();
    static final String indexWordDEQuery = "INSERT INTO IndexWord_DE (indexWord, translationId) VALUES ";
    List<String> indexWordDEValues = new ArrayList<String>();
    static final String indexWordNOQuery = "INSERT INTO IndexWord_NO (indexWord, translationId) VALUES ";
    List<String> indexWordNOValues = new ArrayList<String>();
    static final String stemWordDEQuery = "INSERT INTO StemWord_DE (stemWord, translationId) VALUES ";
    List<String> stemWordDEValues = new ArrayList<String>();
    static final String stemWordNOQuery = "INSERT INTO StemWord_NO (stemWord, translationId) VALUES ";
    List<String> stemWordNOValues = new ArrayList<String>();
    static final String phoneticWordDEQuery = "INSERT INTO PhoneticWord_DE (phoneticWord, baseWord) VALUES ";
    List<String> phoneticWordDEValues = new ArrayList<String>();
    static final String phoneticWordNOQuery = "INSERT INTO PhoneticWord_NO (phoneticWord, baseWord) VALUES ";
    List<String> phoneticWordNOValues = new ArrayList<String>();
    static DbTranslationStore instance;
    
    /**
     * Creates a new instance of DbTranslationStore
     */
    private DbTranslationStore() {
        try {
            conn = DictUtils.dbConnection();
            String lockQuery = "LOCK TABLES "+
                    "Translation WRITE, "+
                    "IndexWord_DE WRITE, "+
                    "IndexWord_NO WRITE, "+
                    "StemWord_DE WRITE, "+
                    "StemWord_NO WRITE, "+
                    "PhoneticWord_DE WRITE, "+
                    "PhoneticWord_NO WRITE ";
            sth = conn.createStatement();
            sth.execute(lockQuery);
            sth.execute("DELETE FROM Translation");
            sth.execute("DELETE FROM IndexWord_DE");
            sth.execute("DELETE FROM IndexWord_NO");
            sth.execute("DELETE FROM StemWord_DE");
            sth.execute("DELETE FROM StemWord_NO");
            sth.execute("DELETE FROM PhoneticWord_DE");
            sth.execute("DELETE FROM PhoneticWord_NO");
        } catch (SQLException se) {
            se.printStackTrace();
        }
    }
    
    synchronized static public DbTranslationStore getInstance() {
        if (instance == null) {
            instance = new DbTranslationStore();
        }
        try {
            if (instance.conn == null || instance.conn.isClosed()) {
                instance = new DbTranslationStore();
            }
        } catch (SQLException se) {
            se.printStackTrace();
        }
        return instance;
    }
    
    @Override
    synchronized public void addTranslation(String word_DE, String article_DE, String other_DE, Collection<String> optional_DE,
            String word_NO, String article_NO, String other_NO, Collection<String> optional_NO,
            String category, String grade) {
        word_DE = word_DE.trim();
        article_DE = article_DE.trim();
        other_DE = other_DE.trim();
        word_NO = word_NO.trim();
        article_NO = article_NO.trim();
        other_NO = other_NO.trim();
        category = category.trim();
        grade = grade.trim();
        grade = grade.equals("") ? "NULL" : grade;
        
        int translationId = ++translationCounter;
        translationValues.add(inBrackets(Utils.join(",", Integer.toString(translationId), inTicks(word_DE), inTicks(article_DE), inTicks(other_DE), inTicks(word_NO), inTicks(article_NO), inTicks(other_NO), inTicks(category), grade)));
        
        Set<String> optionalSet_DE = new HashSet<String>();
        Set<String> stemSet_DE = new HashSet<String>();
        Set<String> phoneticSet_DE = new HashSet<String>();
        for (String str : optional_DE) {
            optionalSet_DE.add(str.toLowerCase(Locale.GERMAN).trim());
        }
        optionalSet_DE.add(word_DE.toLowerCase());
        for (String word : optionalSet_DE) {
            for (String str : Utils.multiplyUmlautWords(word)) {
                if (!str.matches("\\s*")) {
                    indexWordDEValues.add(inBrackets(Utils.join(",", inTicks(str), Integer.toString(translationId))));
                    addStemPhonetics_DE(translationId, word_DE, str, stemSet_DE, phoneticSet_DE);
                }
            }
        }
        
        Set<String> optionalSet_NO = new HashSet<String>();
        for (String str : optional_NO) {
            optionalSet_NO.add(str.toLowerCase(Utils.localeNO).trim());
        }
        optionalSet_NO.add(word_NO.toLowerCase(Utils.localeNO));
        for (String word : optionalSet_NO) {
            for (String str : Utils.multiplyUmlautWords(word)) {
                if (!str.matches("\\s*")) {
                    indexWordNOValues.add(inBrackets(Utils.join(",", inTicks(str), Integer.toString(translationId))));
                    addStemPhonetics_NO(translationId, word_NO, str, stemSet_DE, phoneticSet_DE);
                }
            }
        }
        if (translationCounter % 5000 == 0) writeToDb();
    }
    
    synchronized private boolean writeToDb() {
        if (translationValues.size() == 0) return false;
        try {
//            try {
//                File fh = new File("/tmp/transOut.txt");
//                BufferedWriter bw = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(fh)));
//                bw.write(translationQuery);
//                bw.write(Utils.join(",\n", translationValues));
//                bw.close();
//            } catch (IOException ex) {
//                ex.printStackTrace();
//            }
            sth.execute(translationQuery + Utils.join(",\n",translationValues.toArray()));
            translationValues.clear();
            sth.execute(indexWordDEQuery + Utils.join(",\n", indexWordDEValues.toArray()));
            indexWordDEValues.clear();
            sth.execute(indexWordNOQuery + Utils.join(",\n", indexWordNOValues.toArray()));
            indexWordNOValues.clear();
            sth.execute(stemWordDEQuery + Utils.join(",\n", stemWordDEValues.toArray()));
            stemWordDEValues.clear();
            sth.execute(stemWordNOQuery + Utils.join(",\n", stemWordNOValues.toArray()));
            stemWordNOValues.clear();
            sth.execute(phoneticWordDEQuery + Utils.join(",\n", phoneticWordDEValues.toArray()));
            phoneticWordDEValues.clear();
            sth.execute(phoneticWordNOQuery + Utils.join(",\n", phoneticWordNOValues.toArray()));
            phoneticWordNOValues.clear();
        } catch (SQLException se) {
            se.printStackTrace();
            return false;
        } finally {
            translationValues.clear();
            indexWordDEValues.clear();
            indexWordNOValues.clear();
            stemWordDEValues.clear();
            stemWordNOValues.clear();
            phoneticWordDEValues.clear();
            phoneticWordNOValues.clear();
        }
        return true;
    }
    
    @Override
    synchronized public boolean finish() {
        writeToDb();
        try {
            sth.execute("UNLOCK TABLES");
            
        } catch (SQLException se) {
            se.printStackTrace();
            return false;
        } finally {
            translationValues.clear();
            indexWordDEValues.clear();
            indexWordNOValues.clear();
            stemWordDEValues.clear();
            stemWordNOValues.clear();
            phoneticWordDEValues.clear();
            phoneticWordNOValues.clear();
            translationCounter = 0;
            try { sth.close(); } catch (Exception e) {e.printStackTrace();}
            try { conn.close(); } catch (Exception e) {e.printStackTrace();}
        }
        return true;
    }
    
    private void addStemPhonetics_DE(int id, String word, String str, Set<String> stemSet, Set<String> phoneticSet) {
        org.tartarus.snowball.SnowballProgram stemmer = new org.tartarus.snowball.ext.germanStemmer();
        stemmer.setCurrent(str);
        stemmer.stem();
        String stemmed = stemmer.getCurrent();
        if (stemSet.add(stemmed)) {
            stemWordDEValues.add(inBrackets(Utils.join(",", inTicks(str), Integer.toString(id))));
            AbstractMultiphone multiphone = new info.heinzelnisse.multiphone.De();
            for (String mps : multiphone.multiphone(stemmed)) {
                if (phoneticSet.add(mps)) {
                    phoneticWordDEValues.add(inBrackets(Utils.join(",", inTicks(mps), inTicks(word))));
                }
            }
        }
    }
    private void addStemPhonetics_NO(int id, String word, String str, Set<String> stemSet, Set<String> phoneticSet) {
        org.tartarus.snowball.SnowballProgram stemmer = new org.tartarus.snowball.ext.norwegianStemmer();
        stemmer.setCurrent(str);
        stemmer.stem();
        String stemmed = stemmer.getCurrent();
        if (stemSet.add(stemmed)) {
            stemWordNOValues.add(inBrackets(Utils.join(",", inTicks(str), Integer.toString(id))));
            AbstractMultiphone multiphone = new info.heinzelnisse.multiphone.No();
            for (String mps : multiphone.multiphone(stemmed)) {
                if (phoneticSet.add(mps)) {
                    phoneticWordNOValues.add(inBrackets(Utils.join(",", inTicks(mps), inTicks(word))));
                }
            }
        }
    }
    
    
    static String inTicks(String str) {
        str = str.replaceAll("'","\"");
        return "'" + str + "'";
    }
    static String inBrackets(StringBuilder sb) {
        return sb.insert(0, '(').append(')').toString();
    }
}
