/*
 * DictUtils.java
 *
 * Created on October 18, 2005, 2:27 PM
 * _mem on i5 2410: 100MB 1050#/sec
 * _mysql         : 20MB  940#/sec
 *
 */

package info.heinzelnisse.dictionary;

import info.heinzelnisse.dictionary.dicttables.TranslationDeNo;
import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.sql.Blob;
import java.sql.Connection;
import java.sql.Date;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.sql.DataSource;

/**
 *
 * @author Heiko Klein
 */
public abstract class DictUtils {
    public enum DictLang {DE, NO};
    public enum DictType {dict, trainer, mobile, suggest};
    protected enum Table {
        TRANSLATION ("Translation"),
        INDEX_WORD ("IndexWord"),
        STEM_WORD ("StemWord"),
        PHONETIC_WORD ("PhoneticWord"),
        BOKMAAL_WORD("BokmaalWord"),
        NYNORSK_WORD("NynorskWord");
        
        private final String tableName;
        Table(String name) {
            tableName = name;
        }
        @Override
        public String toString() {
            return tableName;
        }
    }
    private static final Logger logger = Logger.getLogger(DictUtils.class.getName());
    
    abstract public String getTablePrefix();
    
    protected String dbName(Table table) {
        return getTablePrefix() + table.toString();
    }

    protected String dbName(Table table, DictLang deNo) {
        return dbName(table) + "_" + deNo;
    }
    
    /**
     * fetch a connection from the connection-pool
     */
    public static Connection dbConnection() throws SQLException {
        DataSource ds = null;
        try {
            // Obtain our environment naming context
            Context initCtx = new InitialContext();
            Context envCtx = (Context) initCtx.lookup("java:comp/env");
            
            // Look up our data source
            ds = (DataSource) envCtx.lookup("jdbc/HeinzelDB");
        } catch (NamingException ne) {
            System.err.println(ne.toString());
            System.err.println("DataSource not available, using manual connection to heikok_Ordbok");
            try {Class.forName("com.mysql.jdbc.Driver");} catch (ClassNotFoundException cnfe) {cnfe.printStackTrace();}
            return DriverManager.getConnection("jdbc:mysql://localhost/heikok_Ordbok?user=heikok_heikok&amp;characterEncoding=UTF-8&amp;zeroDateTimeBehavior=convertToNull");
        }
        
        // Allocate and use a connection from the pool
        return ds.getConnection();
    }
        
    /**
     * search database for all items with exactly searchitem
     *
     * @return: sorted List of Translation
     * @see dbSearchUniqueId
     */
    abstract public List<? extends Translation> dbSearch(String searchItem, DictLang lang);    

    /**
     * search database for all items with searchitem approximately
     *
     * @return sorted List of Translation
     * @see dbSearchUniqueId
     */
    abstract public List<? extends Translation> dbSearchStart(String searchItem, DictLang lang);
    
    
    /**
     * retrieve a translation collection of the translation-ids
     * 
     * @param ids translation-ids
     * @param language wanted language of the translations (DE = translation from german to norwegian)
     * @return collection of translation
     * 
     */ 
    abstract public Collection<? extends Translation> getTranslations(Collection<Integer> ids, DictUtils.DictLang language);
    
    /**
     * search database for all translation-ids with exactly searchitem
     *
     * @return set of Translation-ids
     */
    protected Set<Integer> dbSearchIdSet(String searchItem, DictLang lang) {
        return DictTables.searchExactIds(searchItem, lang);
    }

    /**
     * search database for all translation-ids with searchitem approximately
     *
     * @return set of Translation-ids
     */
    public Set<Integer> dbSearchStartIdSet(String searchItem, DictLang lang) {
        return DictTables.searchStartIds(searchItem, lang);
    }

    
    /**
     * retrieve a list of words similar (typos, phonetics) to the searchItem
     * 
     * @param searchItem the word inserted
     * @param lang language of the searchItem
     * @return list of similar words, never null
     */
    public List<String> dbSearchSimilarWords(String searchItem, DictLang lang) {
        return DictTables.searchSimilarWords(searchItem, lang);
    }
    

    public Map<String, Suggestion> dbSearchSuggestions(String searchItem) {
        Map<String, Suggestion> retMap = new HashMap<String, Suggestion>();
        searchItem = Utils.cleanWord(searchItem).toLowerCase();
        if (searchItem.length() == 0) {
            return retMap;
        }

        // german index
        Collection<TranslationDeNo> trans = DictTables.getTranslations(DictTables.searchStartIds(searchItem, DictLang.DE));
        for (TranslationDeNo t : trans) {
            String mapWord = t.word_DE + " (D)";
            if (!retMap.containsKey(mapWord)) {
                retMap.put(mapWord, new Suggestion(t.word_DE, DictLang.DE));
            }            
        }
        // norwegian index
        trans = DictTables.getTranslations(DictTables.searchStartIds(searchItem, DictLang.NO));
        for (TranslationDeNo t : trans) {
            String mapWord = t.word_NO + " (N)";
            if (!retMap.containsKey(mapWord)) {
                retMap.put(mapWord, new Suggestion(t.word_NO, DictLang.NO));
            }            
        }
        
        Connection conn = null;
        PreparedStatement statement = null;
        ResultSet rs = null;
        try {
            conn = DictUtils.dbConnection();

            DictLang lang = DictLang.DE;
            // forum keywords
            String queryF = "SELECT keyword FROM ForumKeyword " +
                    " WHERE keyword LIKE ? " +
                    " LIMIT 100 ";
            statement = conn.prepareStatement(queryF);
            statement.setString(1, searchItem + '%');
            rs = statement.executeQuery();
            while (rs.next()) {
                String word = rs.getString("keyword");
                String mapWord = word + " (Forum)";
                if (!retMap.containsKey(mapWord)) {
                    retMap.put(mapWord, new Suggestion(word, lang));
                }
            }
            rs.close();
            statement.close();

            
            // wiki
            String queryW = "SELECT name FROM WikiCurrentPage " +
                    " WHERE name LIKE ? " +
                    " LIMIT 100 ";
            statement = conn.prepareStatement(queryW);
            statement.setString(1, searchItem + '%');
            rs = statement.executeQuery();
            while (rs.next()) {
                String word = rs.getString("name");
                String mapWord = word + " (Wiki)";
                if (!retMap.containsKey(mapWord)) {
                    retMap.put(mapWord, new Suggestion(word, lang));
                }
            }
            rs.close();
            statement.close();


        } catch (SQLException se) {
            se.printStackTrace();
        } finally {
            try {if (rs != null) rs.close();} catch (Exception e) {e.printStackTrace();}
            try {if (statement != null) statement.close();} catch (Exception e) {e.printStackTrace();}
            try {if (conn != null) conn.close();} catch (Exception e) {e.printStackTrace();}
        }

        return retMap;
    }

    private List<String> dbSearchSimilarOrdbankenWords(String searchItem, Table langTable) {
        List<String> wordList = new ArrayList<String>();
        Connection conn = null;
        PreparedStatement statement = null;
        ResultSet rs = null;
        try {
            conn = DictUtils.dbConnection();
            // set limit high, but not infinite
            String query = "SELECT DISTINCT(searchTerm) FROM " + langTable.toString() + " WHERE word = ? ORDER BY searchTerm LIMIT 1000";
            statement = conn.prepareStatement(query);
            statement.setString(1, searchItem);
            rs = statement.executeQuery();
            while (rs.next()) {
                String searchTerm = "";
                try {
                    searchTerm = new String(rs.getBytes("searchTerm"), "utf-8");
                } catch (UnsupportedEncodingException ex) {
                    ex.printStackTrace();
                }
                logger.log(Level.FINE, "found ordbanken-word for {0}:{1}", new Object[]{searchItem, searchTerm});
                wordList.add(searchTerm);
            }
        } catch (SQLException se) {
            se.printStackTrace();
        } finally {
            try {rs.close();} catch (Exception e) {e.printStackTrace();}
            try {statement.close();} catch (Exception e) {e.printStackTrace();}
            try {conn.close();} catch (Exception e) {e.printStackTrace();}
        }
        return wordList;
    }
    
    public List<String> dbSearchSimilarBokmaalWords(String searchItem) {
        return dbSearchSimilarOrdbankenWords(searchItem, Table.BOKMAAL_WORD);
    }

    public List<String> dbSearchSimilarNynorskWords(String searchItem) {
        return dbSearchSimilarOrdbankenWords(searchItem, Table.NYNORSK_WORD);
    }

    
    
    static public String getRandomWord(String lang) {
        Connection conn = null;
        PreparedStatement statement = null;
        ResultSet rs = null;
        try {
            conn = DictUtils.dbConnection();
            String query = "SELECT word_"+lang.toUpperCase()+" AS word FROM Translation ORDER BY rand() LIMIT 1";
            statement = conn.prepareStatement(query);
            rs = statement.executeQuery();
            rs.first();
            String str = rs.getString("word");
            return str;
        } catch (SQLException se) {
            se.printStackTrace();
        } finally {
            try {rs.close();} catch (Exception e) {e.printStackTrace();}
            try {statement.close();} catch (Exception e) {e.printStackTrace();}
            try {conn.close();} catch (Exception e) {e.printStackTrace();}
        }
        return null;
    }

    /**
     * add found/not found words
     * @param conn connection to write wordAccesses to, provided to make sure that connection is available
     * @param wordAccess Array of List with [0] = found words, [1] = not found words
     */
    public void addWordAccess(Connection conn, List<String>[] wordAccess) {
        if (wordAccess != null && wordAccess.length == 2) {
            List<String> foundWords = wordAccess[0];
            List<String> notFoundWords = wordAccess[1];
            assert(foundWords != null);
            assert(notFoundWords != null);
            if (foundWords.size() > 0 || notFoundWords.size() > 0) {
                PreparedStatement statement = null;
                Statement lockStatement = null;
                try {
                    lockStatement = conn.createStatement();
                    lockStatement.execute("LOCK TABLES WordAccess WRITE");
                    String query = "INSERT INTO WordAccess (word, haveAnswer) VALUES (?,?)";
                    statement = conn.prepareStatement(query);
                    for (String word : foundWords) {
                        statement.setString(1, word);
                        statement.setInt(2, 1);
                        statement.addBatch();
                    }
                    for (String word : notFoundWords) {
                        statement.setString(1, word);
                        statement.setInt(2, -1);
                        statement.addBatch();
                    }
                    statement.executeBatch();
                } catch (SQLException se) {
                    se.printStackTrace();
                } finally {
                    try {lockStatement.execute("UNLOCK TABLES");} catch (Exception e) {e.printStackTrace();}
                    try {lockStatement.close();} catch (Exception e) {e.printStackTrace();}
                    try {statement.close();} catch (Exception e) {e.printStackTrace();}
                }

            }
        }
    }
    
    public int clearWordAccessLarger(int amount) {
        int count = -1;
        Connection conn = null;
        PreparedStatement pStatement = null;
        Statement statement = null;
        ResultSet rs = null;
        try {
            conn = DictUtils.dbConnection();
            statement = conn.createStatement();
            statement.execute("LOCK TABLES WordAccess WRITE");
            String query = "SELECT count(*) FROM WordAccess";
            pStatement = conn.prepareStatement(query);
            rs = pStatement.executeQuery();
            rs.first();
            count = rs.getInt(1);
            if (count > amount) {
                statement.execute("DELETE FROM WordAccess");
            }
        } catch (SQLException se) {
            se.printStackTrace();
        } finally {
            try {rs.close();} catch (Exception e) {e.printStackTrace();}
            try {pStatement.close();} catch (Exception e) {e.printStackTrace();}
            try {statement.execute("UNLOCK TABLES");} catch (Exception e) {e.printStackTrace();}
            try {statement.close();} catch (Exception e) {e.printStackTrace();}
            try {conn.close();} catch (Exception e) {e.printStackTrace();}
        }
        return count;
    }

    /**
     * @param conn connection to write wordAccesses to, provided to make sure that connection is available
     * @param usageCounter map with type and count for the current date
     */
    public void incrementUsageCounter(Connection conn, Map<DictUtils.DictType, Integer> usageCounter) {
        assert(conn != null);
        assert(usageCounter != null);
        // create a String, Integer hash, count total new counts
        Map<String, Integer> counter = new HashMap<String, Integer>();
        int total = 0;
        for (Map.Entry<DictType, Integer> entry : usageCounter.entrySet()) {
            DictType type = entry.getKey();
            Integer val = entry.getValue();
            if (val != null) {
                total += val;
                counter.put(type.toString(), val);
            } else {
                counter.put(type.toString(), 0);
            }
        }
        logger.log(Level.FINE, "adding {0} usages to disk", total);
        if (total > 0) {
            // add database counts to counter-hash
            PreparedStatement selectSt = null;
            PreparedStatement updateSt = null;
            PreparedStatement insertSt = null;
            Statement statement = null;
            ResultSet rs = null;
            try {
                statement = conn.createStatement();
                Date date = null;
                rs = statement.executeQuery("SELECT NOW()");
                if (rs.first()) {
                    date = rs.getDate(1);
                }
                rs.close();
                statement.execute("LOCK TABLES DictionaryCounter WRITE");
                // join database and counter counts in counter for date
                selectSt = conn.prepareStatement("SELECT amount, type FROM DictionaryCounter WHERE date = ?");
                selectSt.setDate(1, date);
                rs = selectSt.executeQuery();
                Set<String> existingEntries = new HashSet<String>();
                while (rs.next()) {
                    int amount = rs.getInt("amount");
                    String type = rs.getString("type");
                    existingEntries.add(type);
                    Integer value = counter.get(type);
                    int val = (value == null) ? 0 : value;
                    counter.put(type, val+amount);
                }
                rs.close();
                selectSt.close();
                
                // write counter hash
                updateSt = conn.prepareStatement("UPDATE DictionaryCounter SET amount = ? WHERE date = ? AND type = ?");
                insertSt = conn.prepareStatement("INSERT INTO DictionaryCounter (amount, date, type) VALUES (?,?,?)");
                for (Map.Entry<String,Integer> entry : counter.entrySet()) {
                    String type = entry.getKey();
                    int amount = entry.getValue(); // all values set due to construction of counter
                    PreparedStatement st = (existingEntries.contains(type)) ? updateSt : insertSt;
                    st.setInt(1, amount);
                    st.setDate(2, date);
                    st.setString(3, type);
                    st.execute();
                }
            } catch (SQLException se) {
                se.printStackTrace();
            } finally {
                try {rs.close();} catch (Exception e) {e.printStackTrace();}
                try {selectSt.close();} catch (Exception e) {e.printStackTrace();}
                try {updateSt.close();} catch (Exception e) {e.printStackTrace();}
                try {insertSt.close();} catch (Exception e) {e.printStackTrace();}
                try {statement.execute("UNLOCK TABLES");} catch (Exception e) {e.printStackTrace();}
                try {statement.close();} catch (Exception e) {e.printStackTrace();}
            }
        }
    }
    
    static public String dbLangFlag(DictLang lang) {
        String s = null;
        switch (lang) {
            case DE: s = "DE"; break;
            case NO: s = "NO"; break;
            default:;
        }
        return s;
    }
    static public String dbTransLangFlag(DictLang lang) {
        String s = null;
        switch (lang) {
            case DE: s = "NO"; break;
            case NO: s = "DE"; break;
            default:;
        }
        return s;
    }

    public int countTranslations() {
        logger.fine("Counting translations");
        Connection conn = null;
        PreparedStatement statement = null;
        ResultSet rs = null;
        
        int count = 0;
        try {
            conn = DictUtils.dbConnection();
            String query = "SELECT count(translationId) FROM " + dbName(Table.TRANSLATION);
            statement = conn.prepareStatement(query);
            rs = statement.executeQuery();
            rs.first();
            count = rs.getInt(1);
        } catch (SQLException se) {
            se.printStackTrace();
        } finally {
            try {rs.close();} catch (Exception e) {e.printStackTrace();}
            try {statement.close();} catch (Exception e) {e.printStackTrace();}
            try {conn.close();} catch (Exception e) {e.printStackTrace();}
        }
        return count;
    }
    public int countItems() {
        logger.fine("retrieving countItems");
        Connection conn = null;
        PreparedStatement statement = null;
        ResultSet rs = null;
        
        int count = 0;
        try {
            conn = DictUtils.dbConnection();
            // in mysql 4.1: select count(*) from (select indexWord from IndexWord_DE UNION select indexWord from IndexWord_NO) as dummy
            String query = "SELECT indexWord FROM " + dbName(Table.INDEX_WORD, DictLang.DE) + " UNION SELECT indexWord FROM " + dbName(Table.INDEX_WORD, DictLang.NO);
            statement = conn.prepareStatement(query);
            rs = statement.executeQuery();
            while (rs.next()) {
                count++;
            }
        } catch (SQLException se) {
            se.printStackTrace();
        } finally {
            try {rs.close();} catch (Exception e) {e.printStackTrace();}
            try {statement.close();} catch (Exception e) {e.printStackTrace();}
            try {conn.close();} catch (Exception e) {e.printStackTrace();}
        }
        return count;
    }
    /**
     * return the ratio of successfull searches / all searches in percent
     */
    public int goodTotalRatio() {
        logger.fine("retrieving goodTotalRatio");
        Connection conn = null;
        PreparedStatement statement = null;
        ResultSet rs = null;
        
        int ratio = 0;
        try {
            conn = DictUtils.dbConnection();
            String query = "SELECT count(haveAnswer), haveAnswer FROM WordAccess GROUP BY haveAnswer";
            statement = conn.prepareStatement(query);
            rs = statement.executeQuery();
            int good = 0;
            int bad = 0;
            while (rs.next()) {
                if (rs.getInt(2) > 0) {
                    good += rs.getInt(1);
                } else {
                    bad += rs.getInt(1);
                }
            }
            if ((good+bad) != 0) {
                ratio = (int)(.5 + (100.*good)/(good+bad));
            }
        } catch (SQLException se) {
            se.printStackTrace();
        } finally {
            try {rs.close();} catch (Exception e) {e.printStackTrace();}
            try {statement.close();} catch (Exception e) {e.printStackTrace();}
            try {conn.close();} catch (Exception e) {e.printStackTrace();}
        }
        return ratio;
    }
    
    /**
     * @return int[0] count of dict, int[1] count of trainer
     */
    public int[] countAccessBetween(java.util.Date start, java.util.Date end) {
        Connection conn = null;
        PreparedStatement statement = null;
        ResultSet rs = null;
        
        int[] counts = {0,0};
        try {
            conn = DictUtils.dbConnection();
            String query = "SELECT sum(amount), type FROM DictionaryCounter WHERE date >= ? AND date < ? GROUP BY type";
            statement = conn.prepareStatement(query);
            statement.setTimestamp(1, new Timestamp(start.getTime()));
            statement.setTimestamp(2, new Timestamp(end.getTime()));
            rs = statement.executeQuery();
            while (rs.next()) {
                String type = rs.getString("type");
                if ("suggest".equals(type)) {
                    // ignore currently
                } else if ("trainer".equals(type)) {
                    counts[1] += rs.getInt(1);
                } else {
                    counts[0] += rs.getInt(1);
                }
            }
        } catch (SQLException se) {
            se.printStackTrace();
        } finally {
            try {rs.close();} catch (Exception e) {e.printStackTrace();}
            try {statement.close();} catch (Exception e) {e.printStackTrace();}
            try {conn.close();} catch (Exception e) {e.printStackTrace();}
        }
        logger.log(Level.FINE, "retrieveing countAccess between {0} and {1}: {2}", new Object[]{start, end, counts[0] + counts[1]});
        return counts;
    }

    /**
     * add a heinzelliste table in original (i.e. odt) format to the database
     * @param is the odt-table
     */
    static public void dbStoreOOTable(InputStream is, int size) {
        Connection conn = null;
        PreparedStatement pst = null;
        ResultSet rs = null;
        try {
            conn = DictUtils.dbConnection();
            String query = "INSERT INTO HeinzelListStore(heinzelliste) VALUES(?)";
            pst = conn.prepareStatement(query, Statement.RETURN_GENERATED_KEYS);
            pst.setBlob(1, is);
            int rows = pst.executeUpdate();
            rs = pst.getGeneratedKeys();
            if (rs.next()) {
                int key = rs.getInt(1);
                rs.close();
                pst.close();
                // remove old entries
                pst = conn.prepareStatement("DELETE FROM HeinzelListStore WHERE id != ?");
                pst.setInt(1, key);
                rows = pst.executeUpdate();
                logger.log(Level.FINE, "deleting old tables from heinzellistestore: {0} rows, keep id {1}", new Object[]{rows, key});
            }
        } catch (SQLException se) {
            se.printStackTrace();
        } finally {
            try {rs.close();} catch (Exception e) {e.printStackTrace();}
            try {pst.close();} catch (Exception e) {e.printStackTrace();}
            try {conn.close();} catch (Exception e) {e.printStackTrace();}
        }
        
    }

    /**
     * read the heinzeltable from the database to the ootl store
     * @param ootl 
     */
    static public void dbReadOOTable(OpenOfficeTableLoader ootl) {
        Connection conn = null;
        PreparedStatement pst = null;
        ResultSet rs = null;
        try {
            conn = DictUtils.dbConnection();
            String query = "SELECT id, heinzelliste FROM HeinzelListStore ORDER BY id DESC";
            pst = conn.prepareStatement(query);
            rs = pst.executeQuery();
            if (rs.next()) {
                // there should be only one, or at least, I'm only interested in the latest
                try {
                    Blob table = rs.getBlob(2);
                    ootl.readZipContent(table.getBinaryStream());
                } catch (IOException iox) {
                    logger.warning(iox.toString());
                } finally {
                    ootl.finish();
                }
            }
        } catch (SQLException se) {
            se.printStackTrace();
        } finally {
            try {rs.close();} catch (Exception e) {e.printStackTrace();}
            try {pst.close();} catch (Exception e) {e.printStackTrace();}
            try {conn.close();} catch (Exception e) {e.printStackTrace();}
        }
        
    }

}
