/*
 * HeinzellisteTranslationStore.java
 *
 * Created on Sep 21, 2007, 9:30:52 AM
 *
 */

package info.heinzelnisse.dictionary;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.zip.GZIPOutputStream;

/**
 * write a new heinzelliste to a hardcoded filename
 * 
 * in each directory dicthtml-de-no and dicthtml-no-de run marisa-trie:
 *   ../marisa-0.2.4/tools/marisa-build words.txt > words
 * and zip the directory
 *   zip ../dicthtml-no-de.zip words *.html
 * 
 * @warning the filenames are not locked, the users need to make sure that not
 *          two people upload at the same time!
 *
 * @author hklein
 */
public class HeinzellisteKoboStore implements TranslationStore {

    class KoboTranslation {
        List<String> searchItems;
        Set<String> translations;

        public KoboTranslation() {
            searchItems = new ArrayList<String>();
            translations = new TreeSet<String>();
        }
        
        /**
         * add a translation
         * @param translation 
         */
        public void addTranslation(String word, String article, String other) {
            translations.add(formatTranslation(word, article, other));
        }
        
        public void addSearchItem(String word) {
            if (!searchItems.contains(word)) {
                searchItems.add(word);
            }
        }
        
    }
    
    private Map<String, KoboTranslation> deNo;
    private Map<String, KoboTranslation> noDe;

    public HeinzellisteKoboStore() {
        deNo = new TreeMap<String, KoboTranslation>();
        noDe = new TreeMap<String, KoboTranslation>();
    }
    
    String formatTranslation(String word, String article, String other) {
        StringBuilder sb = new StringBuilder(word);
        if (article != null && !"".equals(article)) {
            sb.append(" <em>(").append(article).append(")</em> ");
        }
        if (other != null && !"".equals(other)) {
            sb.append(" <em>").append(other).append("</em> ");
        }
        return sb.toString();
    }
                
    @Override
    public void addTranslation(String word_DE, String article_DE, String other_DE, Collection<String> optional_DE, String word_NO, String article_NO, String other_NO, Collection<String> optional_NO, String category, String grade) {
        String de = formatTranslation(clean(word_DE), clean(article_DE), clean(other_DE));
        KoboTranslation deKo;
        if (deNo.containsKey(de)) {
            deKo = deNo.get(de);
        } else {
            deKo = new KoboTranslation();
            deNo.put(de, deKo);
        }
        deKo.addTranslation(clean(word_NO), clean(article_NO), clean(other_NO));
        deKo.addSearchItem(clean(word_DE)); 
        // first search item is word itself
        for (String item : optional_DE) {
            deKo.addSearchItem(clean(item));
        }

        String no = formatTranslation(clean(word_NO), clean(article_NO), clean(other_NO));
        KoboTranslation noKo;
        if (noDe.containsKey(no)) {
            noKo = noDe.get(no);
        } else {
            noKo = new KoboTranslation();
            noDe.put(no, noKo);
        }
        noKo.addTranslation(clean(word_DE), clean(article_DE), clean(other_DE));
        // first search item is word itself
        noKo.addSearchItem(clean(word_NO));
        for (String item : optional_NO) {
            noKo.addSearchItem(clean(item));
        }
        return;
    }

    private String clean(String str) {
        if (str == null) return "";
        return str.replaceAll("[\\[\\]\"]", "");
    }

    private String getFileId(String searchItem) {
        String fileId = "11";
        searchItem = searchItem.toLowerCase();
        if (searchItem.matches("^[a-zöäüæøå]$")) {
            fileId = searchItem.substring(0,1) + "a";
        } else if (searchItem.matches("^[a-zöäüæøå]{2}.*")) {
            fileId = searchItem.substring(0, 2);
        }
        return fileId;
    }
    
    private BufferedWriter getWriter(String searchItem, Map<String, BufferedWriter> existing, String dirName) throws IOException {
        String fileId = getFileId(searchItem);
        if (!existing.containsKey(fileId)) {
            BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(new GZIPOutputStream(new FileOutputStream(dirName + fileId + ".html")),"UTF-8"));
            writer.append("<?xml version=\"1.0\" encoding=\"utf-8\"?>\n<html>");
            existing.put(fileId, writer);
            
        }
        return existing.get(fileId);
    }
    
    private void closeWriters(Map<String, BufferedWriter> existing) throws IOException {
        for (BufferedWriter writer : existing.values()) {
            writer.append("</html>");
            writer.close();
        }
    }
    
    private void writeFiles(String dirName, Map<String, KoboTranslation> map) throws IOException {
        Map<String, BufferedWriter> writers = new HashMap<String, BufferedWriter>();
        File wordsFile = new File(dirName + "words.txt");
        OutputStream wos = new FileOutputStream(wordsFile);
        BufferedWriter wosWriter = new BufferedWriter(new OutputStreamWriter(wos, "UTF-8"));

        for (String word : map.keySet()) {
            KoboTranslation trans = map.get(word);
            BufferedWriter writer = getWriter(trans.searchItems.get(0), writers, dirName);
            if (writer != null) {
                writer.append("<w><p>");
                writer.append("<b>").append(word).append("</b><var>");
                for (String o : trans.searchItems) {
                    writer.append("<variant name=\"").append(o).append("\" />");
                }
                writer.append("</var><ul>");
                for (String t : trans.translations) {
                    writer.append("<li>").append(t).append("</li>");
                }
                writer.append("</ul>");
                writer.append("</p></w>");
                // and write the words
                String fileId = getFileId(trans.searchItems.get(0));
                for (String o : trans.searchItems) {
                    if (fileId.equals(getFileId(o))) {
                        if (o.length() > 0)
                            wosWriter.append(o).append("\n");
                    }
                }
            } else {
                System.err.println("no writer for item: " + trans.searchItems.get(0));
            }
        }
        closeWriters(writers);
        wosWriter.close();
    }
            
    
    @Override
    public boolean finish() {
        try {
            //String fileName = java.util.ResourceBundle.getBundle("info/heinzelnisse/dictionary/properties").getString("MobiFileLocation");
            String dirName = "/home/hklein/Programme/Heinzelnisse/KoboDictionary/";
            writeFiles(dirName+"dicthtml-no-de/", noDe);
            writeFiles(dirName+"dicthtml-de-no/", deNo);
        } catch (Exception ex) {
            Logger.getLogger(HeinzellisteKoboStore.class.getName()).log(Level.SEVERE, null, ex);
            return false;
        }
        return true;
    }
}