/*
 * HeinzellisteTranslationStore.java
 *
 * Created on Sep 21, 2007, 9:30:52 AM
 *
 */

package info.heinzelnisse.dictionary;

import info.heinzelnisse.he.HeinzelTable;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Reader;
import java.net.URL;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import org.json.JSONWriter;

/**
 * write a new heinzelliste to a hardcoded filename
 * Run 'kindlegen' on the output-files to obtain a .mobi file, see  http://www.amazon.com/kindleformat/kindlegen
 * 
 * @warning the filenames are not locked, the users need to make sure that not
 *          two people upload at the same time!
 *
 * @author hklein
 */
public class HeinzellisteMobiStore implements TranslationStore {

    class MobiTranslation {
        Set<String> searchItems;
        Set<String> translations;

        public MobiTranslation() {
            searchItems = new TreeSet<String>();
            translations = new TreeSet<String>();
        }
        
        /**
         * add a translation
         * @param translation 
         */
        public void addTranslation(String word, String article, String other) {
            translations.add(formatTranslation(word, article, other));
        }
        
        public void addSearchItem(String word) {
            searchItems.add(word);
        }
        
    }
    
    private Map<String, MobiTranslation> deNo;
    private Map<String, MobiTranslation> noDe;

    public HeinzellisteMobiStore() {
        deNo = new TreeMap<String, MobiTranslation>();
        noDe = new TreeMap<String, MobiTranslation>();
    }
    
    String formatTranslation(String word, String article, String other) {
        StringBuilder sb = new StringBuilder(word);
        if (article != null && !"".equals(article)) {
            sb.append(" <em>(").append(article).append(")</em> ");
        }
        if (other != null && !"".equals(other)) {
            sb.append(" <em>").append(other).append("</em> ");
        }
        return sb.toString();
    }
                
    @Override
    public void addTranslation(String word_DE, String article_DE, String other_DE, Collection<String> optional_DE, String word_NO, String article_NO, String other_NO, Collection<String> optional_NO, String category, String grade) {
        String de = formatTranslation(clean(word_DE), clean(article_DE), clean(other_DE));
        MobiTranslation deMo;
        if (deNo.containsKey(de)) {
            deMo = deNo.get(de);
        } else {
            deMo = new MobiTranslation();
            deNo.put(de, deMo);
        }
        deMo.addTranslation(clean(word_NO), clean(article_NO), clean(other_NO));
        deMo.addSearchItem(clean(word_DE));
        for (String item : optional_DE) {
            deMo.addSearchItem(clean(item));
        }

        String no = formatTranslation(clean(word_NO), clean(article_NO), clean(other_NO));
        MobiTranslation noMo;
        if (noDe.containsKey(no)) {
            noMo = noDe.get(no);
        } else {
            noMo = new MobiTranslation();
            noDe.put(no, noMo);
        }
        noMo.addTranslation(clean(word_DE), clean(article_DE), clean(other_DE));
        noMo.addSearchItem(clean(word_NO));
        for (String item : optional_NO) {
            noMo.addSearchItem(clean(item));
        }
        return;
    }

    private String clean(String str) {
        if (str == null) return "";
        return str.replaceAll("[\\[\\]\"]", "");
    }

    private void writeFile(String fileName, Map<String, MobiTranslation> map) throws IOException {
            File file = new File(fileName + ".new");
            OutputStream os = new FileOutputStream(file);
            BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(os,HeinzelTable.CHARSET));

            if (writer != null) {
                writer.append("<html><head><meta http-equiv=\"Content-Type\" content=\"text/html;charset=iso-8859-1\" /><title>Heinzelnisse</title></head><body><h1>Deutsch Norwegisches W&ouml;rterbuch</h1><h2>Julia Emmerich</h2><mbp:pagebreak />");
                Reader in = new BufferedReader(new InputStreamReader(getClass().getResourceAsStream("cc-by-nc-sa.txt"), "UTF-8"));
                int ch;
                while ((ch = in.read()) > -1) {
                    writer.append((char)ch);
                }
                writer.append("<mbp:pagebreak /><!-- controls --><a onclick=\"index_search()\">Index</a><br /><hr /></center><mbp:pagebreak />\n");
                for (String word : map.keySet()) {
                    MobiTranslation trans = map.get(word);
                    writer.append("<idx:entry>");
                    writer.append("<b><idx:orth>").append(word).append("</idx:orth></b>");
                    for (String o : trans.searchItems) {
                        writer.append("<idx:orth value=\"").append(o).append("\" />");
                    }
                    writer.append("<blockquote>");
                    for (String t : trans.translations) {
                        writer.append(t).append("; ");
                    }
                    writer.append("</blockquote>");
                    writer.append("</idx:entry><hr />\n");
                }
                writer.append("<hr/></body></html>");
                writer.close();
                File finalFile = new File(fileName);
                if (finalFile.exists()) {
                    if (!finalFile.delete()) {
                        // don't care on errors
                    }
                }
                if (!file.renameTo(finalFile)) {
                    Logger.getLogger(HeinzellisteMobiStore.class.getName()).log(Level.SEVERE, "cannot rename filename from: {0} to {1}", new Object[]{file.getAbsolutePath(), finalFile.getAbsolutePath()});
                }
            }
        
    }
            
    
    @Override
    public boolean finish() {
        try {
            //String fileName = java.util.ResourceBundle.getBundle("info/heinzelnisse/dictionary/properties").getString("MobiFileLocation");
            String fileName = "/home/hklein/Programme/Heinzelnisse/KindleDictionary/heinzelnisse_";
            writeFile(fileName+"noDe.html", noDe);
            writeFile(fileName+"deNo.html", deNo);
        } catch (Exception ex) {
            Logger.getLogger(HeinzellisteMobiStore.class.getName()).log(Level.SEVERE, null, ex);
            return false;
        }
        return true;
    }
}