/*
 * OpenOfficeTableLoader.java
 *
 * Created on August 23, 2006, 10:44 AM
 *
 */

package info.heinzelnisse.dictionary;

import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;
import javax.xml.namespace.QName;
import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.XMLReaderFactory;

/**
 *
 * @author Heiko Klein
 */
public class OpenOfficeTableLoader {    
    List<TranslationStore> translationStores = new ArrayList<TranslationStore>();
    
    int rows = 0;
    
    
    /**
     * This class will read an openoffice table to TranslationStores.
     *
     * Usage is as follows:
     *
     * <code>
     *  OpenOfficeTableLoader ootl = new OpenOfficeTableLoader();
     *  ootl.addTranslationStore(DbTranslationStore.getInstance());
     *  //... add more TranslationStores
     *  ootl.readZipContent(InputStream stream); // stream should contain a zip file
     *  //... add more contents
     *  ootl.finish(); // finish the translationStores
     * </code>
     */
    public OpenOfficeTableLoader() {
    }
    
    public void addTranslationStore(TranslationStore tl) {
        translationStores.add(tl);
    }
    
    public int getRowNumber() {
        return rows;
    }
    
    public void readZipURL(URL url) throws IOException {
        readZipContent(url.openStream());
    }
    
    public void readZipContent(InputStream stream) throws IOException {
        ZipInputStream zis = new ZipInputStream(stream);
        ZipEntry ze = zis.getNextEntry();
        while (ze != null) {
            if (ze.getName().equals("content.xml")) {
                readContent(zis);
                break;
            }
            zis.closeEntry();
            ze = zis.getNextEntry();
        }
    }
    /**
     * finish the translation stores (i.e. commit, close streams...)
     */
    public void finish() {
        for (TranslationStore tl : translationStores) {
            tl.finish();
        }
    }
    
    private void readContent(InputStream is) {
        try {
            XMLReader xr = XMLReaderFactory.createXMLReader();
            xr.setFeature("http://xml.org/sax/features/validation", false);
            
            SAXReadXMLHandler fxh = new SAXReadXMLHandler();
            
            xr.setContentHandler(fxh);
            xr.setErrorHandler(fxh);
            xr.parse(new InputSource(is));
            rows = fxh.rows;
        } catch (SAXException sx) {
            sx.printStackTrace();
        } catch (IOException iox) {
            iox.printStackTrace();
        }
        
    }
    
    private class SAXReadXMLHandler extends org.xml.sax.helpers.DefaultHandler {

        final String nameSpaceURI = "urn:oasis:names:tc:opendocument:xmlns:table:1.0";
        final QName qRowName = new QName(nameSpaceURI, "table-row");
        final QName qCellName = new QName(nameSpaceURI, "table-cell");
        final String qCollsRepeated = "table:number-columns-repeated";
        final QName qCellTextName = new QName("urn:oasis:names:tc:opendocument:xmlns:text:1.0", "p");

        int inRow = 0;
        int inCell = 0;
        int inCellText = 0;
        public int rows = 0;
        public int cells = 0;
        String cellText = "";
        List<String> rowElements = new ArrayList<String>(16);
        
        private void addRowElementsToTranslations() {
            String word_NO = getRowElement(0);
            if (word_NO.startsWith("#")) return;
            String article_NO = getRowElement(1);
            String opt_NO = getRowElement(2);
            String other_NO = getRowElement(3);
            String word_DE = getRowElement(4);
            String article_DE = getRowElement(5);
            String opt_DE = getRowElement(6);
            String other_DE = getRowElement(7);
            String category = getRowElement(8);
            if (word_DE.equals("") || word_NO.equals("")) return;
            String grade = "";
            List<String> optional_DE = Arrays.asList(opt_DE.split(";"));
            List<String> optional_NO = Arrays.asList(opt_NO.split(";"));
            for (TranslationStore tl : translationStores) {
                tl.addTranslation(word_DE, article_DE, other_DE, optional_DE, word_NO, article_NO, other_NO, optional_NO, category, grade);
            }
        }
        
        private String getRowElement(int i) {
            if (i >= rowElements.size()) {
                return "";
            }
            String element = rowElements.get(i);
            if (element == null) {
                return "";
            }
            return element;
        }
        
        @Override
        public void startElement(String uri, String name, String qName, Attributes atts) {
            QName qqName = new QName(uri,name);
            if (isRow(qqName)) {
                inRow++;
            } else if ((inRow > 0) && isCell(qqName)) {
                inCell++;
                String repeat = atts.getValue(qCollsRepeated);
                if (repeat != null && !repeat.equals("")) {
                    int iRepeat = Integer.parseInt(repeat);
                    if (iRepeat < 10) { // end of row is about 200 elements
                        for (int i = 1; i < iRepeat; i++) { // 1 will be added regularly
                            rowElements.add("");
                        }
                    }
                }
                cellText = "";
            } else if ((inCell > 0) && isCellText(qqName)) {
                inCellText++;
            }
        }
        @Override
        public void endElement(String uri, String name, String qName) {
            QName qqName = new QName(uri,name);
//            System.err.println(qqName);
            if (isRow(qqName)) {
                if (rowElements.size() > 5) {
                    addRowElementsToTranslations();
                }
                inRow--;
                if (inRow == 0) {
                    // TODO: write back row
                    rows++;
                    rowElements = new ArrayList<String>(16);
                }
            } else if ((inRow > 0) && isCell(qqName)) {
                cells++;
                inCell--;
                rowElements.add(cellText);
            } else if ((inCell > 0) && isCellText(qqName)) {
                inCellText--;
            }
        }
        
        @Override
        public void characters(char[] ch, int start, int length) throws SAXException {
            if (inCellText > 0) {
                cellText += new String(ch, start, length);
            }
        }
        public boolean isRow(QName qname) {
//            if (qname.toString().contains("row")) {
//                System.err.println(qname.toString());
//                System.err.println(qRowName.toString());
//                System.err.println("");
//            }
            if (qname.equals(qRowName)) {
                return true;
            } else {
                return false;
            }
        }
        public boolean isCell(QName qname) {
            if (qname.equals(qCellName)) {
                return true;
            } else {
                return false;
            }
        }
        public boolean isCellText(QName qname) {
            if (qname.equals(qCellTextName)) {
                return true;
            } else {
                return false;
            }
        }
        
    }
    
}
