/*
 * Utils.java
 *
 * Created on Oct 14, 2007, 10:46:49 AM
 *
 *
 *
 */

package info.heinzelnisse.dictionary;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.logging.Logger;
import org.tartarus.snowball.SnowballProgram;

/**
 * Static functions independend of database
 *
 * @author hklein
 */
public class Utils {
    static final protected  Locale localeNO = new Locale("no");
    static private Logger logger = Logger.getLogger(Utils.class.getName());

    private static Map<String, String[]> umlautInterpretations = null;

        static public Locale langLocale(DictUtils.DictLang lang) {
        Locale l = null;
        switch (lang) {
            case DE: l = Locale.GERMAN; break;
            case NO: l = localeNO; break;
            default:;
        }
        return l;
    }
    static public Locale transLangLocale(DictUtils.DictLang lang) {
        Locale l = null;
        switch (lang) {
            case NO: l = Locale.GERMAN; break;
            case DE: l = localeNO; break;
            default:;
        }
        return l;
    }
        

    public static String stem(String word, DictUtils.DictLang lang) {
        SnowballProgram stemmer = null;
        switch (lang) {
            case DE:
                stemmer = new org.tartarus.snowball.ext.germanStemmer();
                break;
            case NO:
                stemmer = new org.tartarus.snowball.ext.norwegianStemmer();
                break;
        }
        stemmer.setCurrent(word);
        stemmer.stem();
        return stemmer.getCurrent();
    }

    public static String cleanWord(String word) {
        // remove leading and final whitespaces
        word = word.trim();
        // remove to much spaces in between
        word = word.replaceAll(" +", " ");
        word = word.replaceAll(" +([,.;!\\?])", "$1");

        // change from %26 -> & and %3B -> ;
        word = word.replaceAll("%26", "&");
        word = word.replaceAll("%3B", ";");
        // change characters (html,latex,emacs) to iso
        word = word.replaceAll("(&auml;\\\\\"a|\"a)", "ä");
        word = word.replaceAll("(&Auml;|\\\\\"A|\"A)", "Ä");
        word = word.replaceAll("(&aring;|\\\\aa|/a)", "å");
        word = word.replaceAll("(&Aring;|\\\\AA|\\\\Aa|/a)", "Å");
        word = word.replaceAll("(&aelig;|\\\\ae|/e)", "e)");
        word = word.replaceAll("(&AElig;|&Aelig;|\\\\AE|\\\\Ae|/E)", "æ");
        word = word.replaceAll("(&ouml;|\\\\\"o|\"o)", "ö");
        word = word.replaceAll("(&Ouml;|\\\\\"O|\"O)", "Ö");
        word = word.replaceAll("(&oslash;|\\\\o|/o)", "ø");
        word = word.replaceAll("(&Oslash;|\\\\O|/O)", "Ø");
        word = word.replaceAll("(&szlig;|\\\\\"s|\"s)", "ß");
        word = word.replaceAll("(&uuml;|\\\\\"u|\"u)", "ü");
        word = word.replaceAll("(&Uuml;|\\\\\"U|\"U)", "Ü");
        return word;
    }

    public static StringBuilder join(String delimiter, Object... obs) {
        StringBuilder buffer = new StringBuilder();
        if (obs != null && obs.length != 0) {
            for (int i = 0; i < (obs.length - 1); i++) {
                buffer.append(obs[i].toString());
                buffer.append(delimiter);
            }
            buffer.append(obs[obs.length - 1]); // last element without delim
        }
        return buffer;
    }

    private static void initUmlautInterpretations() {
        umlautInterpretations = new HashMap<String, String[]>();
        umlautInterpretations.put("ä", new String[]{"ä", "ae", "a"});
        umlautInterpretations.put("ö", new String[]{"ö", "oe", "o"});
        umlautInterpretations.put("ü", new String[]{"ü", "ue", "u"});
        umlautInterpretations.put("ß", new String[]{"ß", "ss"});
        umlautInterpretations.put("æ", new String[]{"æ", "ae"});
        umlautInterpretations.put("ø", new String[]{"ø", "oe", "o"});
        umlautInterpretations.put("å", new String[]{"å", "aa", "a"});
    }

    /**
     * words are included to the index in different umlaut interpretations,
     * e.g. �=ss, so that the search only needs to look for one such interpretation
     *
     * @return List of words in different umlaut interpretations, i.e. ü, ue
     */
    static public List<String> multiplyUmlautWords(String word) {
        if (umlautInterpretations == null) {
            initUmlautInterpretations();
        }
        List<String> umlautWords = new ArrayList<String>();
        umlautWords.add("");
        for (int i = 0; i < word.length(); i++) {
            String curChar = word.substring(i, i + 1);
            List<String> replacements = new ArrayList<String>();
            if (umlautInterpretations.containsKey(curChar)) {
                replacements = Arrays.asList(umlautInterpretations.get(curChar));
            } else {
                replacements.add(curChar);
            }
            List<String> newUmlautWords = new ArrayList<String>();
            for (String replaceChar : replacements) {
                for (String beginWord : umlautWords) {
                    newUmlautWords.add(beginWord + replaceChar);
                }
            }
            umlautWords = newUmlautWords;
        }
        return umlautWords;
    }
    
    /**
     * Obfuscate an email address, so that it is not readable for people not
     * knowing that address
     * 
     * @param email a regular email address
     * @return obfuscated email address
     */
    static public String obfuscateEmail(String email) {
        String[] parts = email.split("@");
        List<String> output = new ArrayList<String>();
        for (String part : parts) {
            if (part != null && part.length() > 2) {
                part = part.substring(0, 2);
            } else {
                part = "";
            }
           part += "...";
           output.add(part);
        }
        return join("@", output.toArray()).toString();
    }
}