/*
 * Question.java
 *
 * Created on November 9, 2005, 10:15 AM
 *
 */

package info.heinzelnisse.forum;
import info.heinzelnisse.dictionary.DictUtils;
import info.heinzelnisse.dictionary.Utils;
import info.heinzelnisse.he.HeinzelTable;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.sql.Types;
import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collections;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 *
 * @author Heiko Klein
 */
public class Question {
    private int id = 0;
    private String text;
    private String originalText;
    private Date timestamp;
    private String address;
    private int userId = 0;
    private List<Answer> answers;
    private List<String> keywords;
    private User user;
    
    /** Creates a new instance of Question */
    public Question() {
    }
    public Question(int id, String text, Date timestamp, String address, int userId) {
        this.id = id;
        this.text = text;
        this.timestamp = timestamp;
        this.address = address;
        this.userId = userId;
    }

    public Question(int id, String text, String originalText, Date timestamp, String address, int userId) {
        this.id = id;
        this.text = text;
        this.originalText = originalText;
        this.timestamp = timestamp;
        this.address = address;
        this.userId = userId;
    }

    static public List<Question> ofMonth(int month, int year) {
        Calendar cal = GregorianCalendar.getInstance();
        cal.set(year, month-1, 1, 0, 0,0);
        Date startDate = cal.getTime();
        cal.add(Calendar.MONTH, 1);
        Date endDate = cal.getTime();
        return ofDate(startDate, endDate, -1);
    }
    
    static public List<Question> ofDate(Date date, int days, int minCount) {
        if (days <= 0) {
            days = 1;
        }
        Calendar cal = GregorianCalendar.getInstance();
        cal.setTime(date);
        // calculate the end-date
        cal.set(Calendar.HOUR_OF_DAY, 23);
        cal.set(Calendar.MINUTE, 59);
        cal.set(Calendar.SECOND, 59);
        Date endDate = new Date(cal.getTimeInMillis());
        cal.add(Calendar.DAY_OF_YEAR, -1 * days);
        cal.set(Calendar.HOUR_OF_DAY, 0);
        cal.set(Calendar.MINUTE, 0);
        cal.set(Calendar.SECOND, 0);
        Date startDate = new Date(cal.getTimeInMillis());
     
        return ofDate(startDate, endDate, minCount);
    }
    
    /**
     *
     * @param startDate
     * @param endDate
     * @param minCount if minCount > 0, use startDate optional and use either
     *                 all elements since startDate but at list minCount elements,
     *                 eventually older
     * @return
     */
    static public List<Question> ofDate(Date startDate, Date endDate, int minCount) {
        List<Question> questions = new ArrayList<Question>();
        Connection conn = null;
        PreparedStatement statement = null;
        ResultSet rs = null;
        
        try {
            conn = DictUtils.dbConnection();
            String query =
                    "SELECT forumQuestion, text, originalText, timestamp, address, forumUser "+
                    "FROM ForumQuestion "+
                    "WHERE  timestamp > ? AND timestamp <= ? "+
                    "ORDER BY timestamp DESC ";
            statement = conn.prepareStatement(query);
            statement.setTimestamp(1,new Timestamp(startDate.getTime()));
            statement.setTimestamp(2,new Timestamp(endDate.getTime()));
            rs = statement.executeQuery();
            while (rs.next()) {
                questions.add(new Question(rs.getInt("forumQuestion"), rs.getString("text"), rs.getString("originalText"), rs.getTimestamp("timestamp"), rs.getString("address"), rs.getInt("forumUser")));
            }
            int repeat = 100;
            while (repeat > 0 && minCount > 0 && questions.size() < minCount) {
                repeat--;
                rs.close();
                endDate = startDate;
                Calendar cal = GregorianCalendar.getInstance();
                cal.setTime(startDate);
                cal.add(Calendar.DAY_OF_YEAR, -7);
                startDate = new Date(cal.getTimeInMillis());
                statement.setTimestamp(1, new Timestamp(startDate.getTime()));
                statement.setTimestamp(2,new Timestamp(endDate.getTime()));
                rs = statement.executeQuery();
                while (rs.next() && questions.size() < minCount) {
                    questions.add(new Question(rs.getInt("forumQuestion"), rs.getString("text"), rs.getString("originalText"), rs.getTimestamp("timestamp"), rs.getString("address"), rs.getInt("forumUser")));
                }                
            }
        } catch (SQLException se) {
            se.printStackTrace();
        } finally {
            try {rs.close();} catch (Exception e) {e.printStackTrace();}
            try {statement.close();} catch (Exception e) {e.printStackTrace();}
            try {conn.close();} catch (Exception e) {e.printStackTrace();}
        }
        return questions;
    }
    static public List<Integer> idOfDate(Date startDate, Date endDate) {
        List<Integer> qIds = new ArrayList<Integer>();
        Connection conn = null;
        PreparedStatement statement = null;
        ResultSet rs = null;
        
        try {
            conn = DictUtils.dbConnection();
            String query =
                    "SELECT forumQuestion "+
                    "FROM ForumQuestion "+
                    "WHERE  timestamp > ? AND timestamp <= ? "+
                    "ORDER BY timestamp DESC ";
            statement = conn.prepareStatement(query);
            statement.setTimestamp(1,new Timestamp(startDate.getTime()));
            statement.setTimestamp(2,new Timestamp(endDate.getTime()));
            rs = statement.executeQuery();
            while (rs.next()) {
                qIds.add(rs.getInt("forumQuestion"));
            }
        } catch (SQLException se) {
            se.printStackTrace();
        } finally {
            try {rs.close();} catch (Exception e) {e.printStackTrace();}
            try {statement.close();} catch (Exception e) {e.printStackTrace();}
            try {conn.close();} catch (Exception e) {e.printStackTrace();}
        }
        return qIds;        
    }
    
    static public Map<Date,int[]> getArchivedMonths() {
        Map<Date,int[]> monthQuestions = new HashMap<Date,int[]>();
        Connection conn = null;
        PreparedStatement statement = null;
        ResultSet rs = null;
        try {
            conn = DictUtils.dbConnection();
            DateFormat dateformat = new SimpleDateFormat("yyyy-MM-dd");
            // returns month as YYYY-MM-01
            String query = 
                    "SELECT DATE_FORMAT(q.timestamp, '%Y-%m-01') as month, COUNT(DISTINCT(q.forumQuestion)) as qAmount, COUNT(a.forumAnswer) as aAmount "+
                    "FROM ForumAnswer a, ForumQuestion q "+
                    "WHERE q.forumQuestion = a.forumQuestion "+
                    "GROUP BY month ORDER BY month";
            statement = conn.prepareStatement(query);
            rs = statement.executeQuery();
            while (rs.next()) {
                String month = rs.getString("month");
                int qAmount = rs.getInt("qAmount");
                int aAmount = rs.getInt("aAmount");
                int[] questionAnswer = new int[] {qAmount, aAmount};
                Date date = dateformat.parse(month);
                monthQuestions.put(date, questionAnswer);
            }
            
        } catch (SQLException se) {
            se.printStackTrace();
        } catch (ParseException pe) {
            pe.printStackTrace();
        } finally {
            try {rs.close();} catch (Exception e) {e.printStackTrace();}
            try {statement.close();} catch (Exception e) {e.printStackTrace();}
            try {conn.close();} catch (Exception e) {e.printStackTrace();}
        }
        return monthQuestions;
    }
    
    static public List<Question> searchByText(String text) {
        List<Question> questions = new ArrayList<Question>();
        Connection conn = null;
        PreparedStatement statement = null;
        ResultSet rs = null;
        try {
            conn = DictUtils.dbConnection();
            Map<Integer, QuestionRank> ranks = new HashMap<Integer, QuestionRank>();
            String query =
                    "SELECT forumQuestion, text, originalText, timestamp, address, forumUser, MATCH(text) AGAINST (?) as rank "+
                    "FROM ForumQuestion "+
                    "WHERE MATCH(text) AGAINST (?) ";
            statement = conn.prepareStatement(query);
            statement.setString(1,text);
            statement.setString(2,text);
            rs = statement.executeQuery();
            while (rs.next()) {
                ranks.put(rs.getInt("forumQuestion"), new QuestionRank(new Question(rs.getInt("forumQuestion"), rs.getString("text"), rs.getString("originalText"), rs.getTimestamp("timestamp"), rs.getString("address"), rs.getInt("forumUser")), rs.getDouble("rank")));
            }
            rs.close();
            statement.close();
            query =
                    "SELECT q.forumQuestion, q.text, q.originalText, q.timestamp, q.address, q.forumUser, MATCH(ForumAnswer.text) AGAINST (?) as rank "+
                    "FROM ForumAnswer, ForumQuestion q "+
                    "WHERE ForumAnswer.forumQuestion = q.forumQuestion "+
                    "  AND MATCH(ForumAnswer.text) AGAINST (?) ";
            statement = conn.prepareStatement(query);
            statement.setString(1,text);
            statement.setString(2, text);
            rs = statement.executeQuery();
            while (rs.next()) {
                double rank = rs.getDouble("rank");
                int forumQuestion = rs.getInt("forumQuestion");
                QuestionRank qr = ranks.get(forumQuestion);
                if (qr != null) {
                    if (qr.rank < rank) {
                        qr.rank = rank;
                    }
                } else {
                    ranks.put(forumQuestion, new QuestionRank(new Question(rs.getInt("forumQuestion"), rs.getString("text"), rs.getString("originalText"), rs.getTimestamp("timestamp"), rs.getString("address"), rs.getInt("forumUser")), rank));
                }
            }
            List<QuestionRank> qrs = new ArrayList<QuestionRank>(ranks.values());
            Collections.sort(qrs);
            for(QuestionRank qr : qrs) {
                questions.add(qr.question);
            }
        } catch (SQLException se) {
            se.printStackTrace();
        } finally {
            try {rs.close();} catch (Exception e) {e.printStackTrace();}
            try {statement.close();} catch (Exception e) {e.printStackTrace();}
            try {conn.close();} catch (Exception e) {e.printStackTrace();}
        }
        return questions;
    }
    
    /**
     * @return null on failure
     */
    static public Question searchByExactText(String originalText) {
        Question question = null;
        Connection conn = null;
        PreparedStatement statement = null;
        ResultSet rs = null;
        try {
            conn = DictUtils.dbConnection();
            String query =
                    "SELECT forumQuestion, text, originalText, timestamp, address, forumUser "+
                    "FROM ForumQuestion "+
                    "WHERE  originalText = ? ";
            statement = conn.prepareStatement(query);
            statement.setString(1,originalText);
            rs = statement.executeQuery();
            while (rs.next()) {
                // should be only one, but who cares
                question = new Question(rs.getInt("forumQuestion"), rs.getString("text"), rs.getString("originalText"), rs.getTimestamp("timestamp"), rs.getString("address"), rs.getInt("forumUser"));
            }
        } catch (SQLException se) {
            se.printStackTrace();
        } finally {
            try {rs.close();} catch (Exception e) {e.printStackTrace();}
            try {statement.close();} catch (Exception e) {e.printStackTrace();}
            try {conn.close();} catch (Exception e) {e.printStackTrace();}
        }
        return question;
    }
    
    static public Question searchById(int id) {
        Question question = null;
        Connection conn = null;
        PreparedStatement statement = null;
        ResultSet rs = null;
        try {
            conn = DictUtils.dbConnection();
            String query =
                    "SELECT forumQuestion, text, originalText, timestamp, address, forumUser "+
                    "FROM ForumQuestion "+
                    "WHERE  forumQuestion = ? ";
            statement = conn.prepareStatement(query);
            statement.setInt(1,id);
            rs = statement.executeQuery();
            while (rs.next()) {
                // should be only one, but who cares
                question = new Question(rs.getInt("forumQuestion"), rs.getString("text"), rs.getString("originalText"), rs.getTimestamp("timestamp"), rs.getString("address"), rs.getInt("forumUser"));
            }
        } catch (SQLException se) {
            se.printStackTrace();
        } finally {
            try {rs.close();} catch (Exception e) {e.printStackTrace();}
            try {statement.close();} catch (Exception e) {e.printStackTrace();}
            try {conn.close();} catch (Exception e) {e.printStackTrace();}
        }
        return question;
    }

    /**
     * search the forum by keywords
     * 
     * @param searchItem clean searchItem directly put into the sql-query
     * @return list of questions
     */
    public static List<Question> searchByKeyword(String searchItem) {
        List<Question> questions = new ArrayList<Question>();
        Connection conn = null;
        PreparedStatement statement = null;
        ResultSet rs = null;
        
        try {
            conn = DictUtils.dbConnection();
            String query =
                    "SELECT DISTINCT(ForumQuestion.forumQuestion), text, originalText, timestamp, address, forumUser "+
                    "FROM ForumQuestion, ForumKeyword "+
                    "WHERE ForumQuestion.forumQuestion = ForumKeyword.forumQuestion " +
                    "AND ForumKeyword.keyword = ? "+
                    "ORDER BY timestamp DESC ";
            statement = conn.prepareStatement(query);
            statement.setString(1, searchItem);
            rs = statement.executeQuery();
            while (rs.next()) {
                questions.add(new Question(rs.getInt("forumQuestion"), rs.getString("text"), rs.getString("originalText"), rs.getTimestamp("timestamp"), rs.getString("address"), rs.getInt("forumUser")));
            }
        } catch (SQLException se) {
            se.printStackTrace();
        } finally {
            try {rs.close();} catch (Exception e) {e.printStackTrace();}
            try {statement.close();} catch (Exception e) {e.printStackTrace();}
            try {conn.close();} catch (Exception e) {e.printStackTrace();}
        }
        return questions;
    }
/**
     * search the forum by user
     *
     * @param searchItem clean searchItem directly put into the sql-query
     * @return list of questions
     */
    public static List<Question> searchByForumUser(int forumUser) {
        List<Question> questions = new ArrayList<Question>();
        Connection conn = null;
        PreparedStatement statement = null;
        ResultSet rs = null;

        try {
            conn = DictUtils.dbConnection();
            String query =
                    "SELECT DISTINCT(q.forumQuestion), q.text, q.originalText, q.timestamp, q.address, q.forumUser "+
                    "FROM ForumQuestion q LEFT JOIN ForumAnswer a "+
                    "  ON q.forumQuestion = a.forumQuestion " +
                    "WHERE q.forumUser = ? " +
                    "   OR a.forumUser = ? " +
                    "ORDER BY q.timestamp DESC ";
            statement = conn.prepareStatement(query);
            statement.setInt(1, forumUser);
            statement.setInt(2, forumUser);
            rs = statement.executeQuery();
            while (rs.next()) {
                questions.add(new Question(rs.getInt("q.forumQuestion"), rs.getString("q.text"), rs.getString("q.originalText"), rs.getTimestamp("q.timestamp"), rs.getString("q.address"), rs.getInt("q.forumUser")));
            }
        } catch (SQLException se) {
            se.printStackTrace();
        } finally {
            try {rs.close();} catch (Exception e) {e.printStackTrace();}
            try {statement.close();} catch (Exception e) {e.printStackTrace();}
            try {conn.close();} catch (Exception e) {e.printStackTrace();}
        }
        return questions;
    }

    static public Question create(String text, String originalText, String keywords, Date timestamp, String address, int userId) {
        Question question = null;
        Connection conn = null;
        PreparedStatement statement = null;
        ResultSet rs = null;
        try {
            conn = DictUtils.dbConnection();
            String query =
                    "INSERT INTO ForumQuestion (text, originalText, timestamp, address, forumUser) "+
                    "VALUES (?,?,?,?,?) ";
            statement = conn.prepareStatement(query, PreparedStatement.RETURN_GENERATED_KEYS);
            statement.setString(1, text);
            statement.setString(2, originalText);
            statement.setTimestamp(3, new Timestamp(timestamp.getTime()));
            statement.setString(4, address);
            if (userId != 0) {
                statement.setInt(5,  userId);
            } else {
                //statement.setNull(5, Types.INTEGER);
                throw new SQLException("no user defined for new question");
            }
            statement.execute();
            rs = statement.getGeneratedKeys();
            if (rs.next()) {
                question = new Question(rs.getInt(1), text, originalText, timestamp, address, userId);
            } else {
                throw new SQLException("cannot generate new question");
            }
        } catch (SQLException se) {
            se.printStackTrace();
        } finally {
            try {rs.close();} catch (Exception e) {e.printStackTrace();}
            try {statement.close();} catch (Exception e) {e.printStackTrace();}
            try {conn.close();} catch (Exception e) {e.printStackTrace();}
        }
        if (question != null) {
            question.addKeywords(keywords);
        }
        return question;
    }
    
    public int getId() {
        return id;
    }
    
    /**
     * Temporary set this text
     */
    public void setText(String text) {
        this.text = text;
    }
    
    public String getText() {
        return text;
    }
    
    /**
     * Temporary set this originalText
     */
    public void setOriginalText(String originalText) {
        this.originalText = originalText;
    }
    
    public String getOriginalText() {
        return originalText;
    }
    
    public Date getTimestamp() {
        return timestamp;
    }
    
    public String getAddress() {
        return address;
    }
    
    public int getUserId() {
        return userId;
    }
    
    public User getUser() {
        if (user == null) {
            user = new User(userId);
        }
        return user;
    }
    
    public List<Answer> getAnswers() {
        if (answers != null) {
            return answers;
        }
        answers = Answer.getAnswers(this);
        return answers;
    }
    
    /**
     * 
     * @param str
     * @see addKeywords
     */
    void addKeywords(String str) {
        int added = addKeywords(id, str);
        if (added > 0) {
        // clear current keyword cache
            this.keywords = null;
        }
    }
    
    /**
     * add keywords to questionId
     * @param id question id
     * @param str comma-separated keywords
     * @return keywords added
     */
    static public int addKeywords(int id, String str) {
        if (str == null) {return 0;}
        Set<String> myKeywords = new HashSet<String>();
        for (String substr : str.split(",")) {
            String keyword = substr.trim();
            if (!keyword.equals("")) {
                myKeywords.add(keyword);
            }
        }
        if (myKeywords.size() < 0) {
            return 0;
        }

        // add to db
        Connection conn = null;
        PreparedStatement statement = null;
        try {
            conn = DictUtils.dbConnection();
            String query = "INSERT INTO ForumKeyword (keyword, forumQuestion) VALUES (?, ?)";
            statement = conn.prepareStatement(query);
            for (String keyword : myKeywords) {
                statement.setString(1, keyword);
                statement.setInt(2, id);
                statement.execute();
            }
        } catch (SQLException se) {
            se.printStackTrace();
        } finally {
            try {statement.close();} catch (Exception e) {e.printStackTrace();}
            try {conn.close();} catch (Exception e) {e.printStackTrace();}
        }
        return myKeywords.size();
    }

    /**
     * remove all keywords for a question
     *
     * @param id question
     */
    static public void removeKeywords(int id) {
        Connection conn = null;
        PreparedStatement statement = null;
        try {
            conn = DictUtils.dbConnection();
            String query = "DELETE FROM ForumKeyword WHERE forumQuestion = ?";
            statement = conn.prepareStatement(query);
            statement.setInt(1,id);
            statement.execute();
        } catch (SQLException se) {
            se.printStackTrace();
        } finally {
            try {statement.close();} catch (Exception e) {e.printStackTrace();}
            try {conn.close();} catch (Exception e) {e.printStackTrace();}
        }
    }

    /**
     * replace all keywords for the question with the supplied ones
     * calls internally @see{removeKeywords} and @see{addKeywords}
     * 
     * @param id
     * @param str comma-separated list of strings
     * @return keywords added
     */
    static public int replaceKeywords(int id, String str) {
        removeKeywords(id);
        return addKeywords(id, str);
    }

    public List<String> getKeywords() {
        if (keywords != null) {
            return keywords;
        }
        keywords = new ArrayList<String>(4);
        Connection conn = null;
        PreparedStatement statement = null;
        ResultSet rs = null;
        try {
            conn = DictUtils.dbConnection();
            String query = "SELECT DISTINCT(keyword) FROM ForumKeyword WHERE forumQuestion = ?";
            statement = conn.prepareStatement(query);
            statement.setInt(1, id);
            rs = statement.executeQuery();
            while (rs.next()) {
                keywords.add(rs.getString("keyword"));
            }
        } catch (SQLException se) {
            se.printStackTrace();
        } finally {
            try {rs.close();} catch (Exception e) {e.printStackTrace();}
            try {statement.close();} catch (Exception e) {e.printStackTrace();}
            try {conn.close();} catch (Exception e) {e.printStackTrace();}
        }
        return keywords;
    }
    
    public String getKeywordString() {
        List<String> myKeywords = getKeywords();
        if (myKeywords == null) return "";
        Collections.sort(myKeywords, HeinzelTable.getDeNoCollator());
        return Utils.join(", ", myKeywords.toArray()).toString();
    }
    
    /**
     * remove this question and all its answers and keywords from the database
     */
    public void removeFromDb() {
        if (id != 0) {
            Connection conn = null;
            PreparedStatement statement = null;
            try {
                conn = DictUtils.dbConnection();
                String query = "DELETE FROM ForumAnswer WHERE forumQuestion = ?";
                statement = conn.prepareStatement(query);
                statement.setInt(1,id);
                statement.execute();
                statement.close();
                query = "DELETE FROM ForumKeyword WHERE forumQuestion = ?";
                statement = conn.prepareStatement(query);
                statement.setInt(1,id);
                statement.execute();
                statement.close();
                query = "DELETE FROM ForumQuestion WHERE forumQuestion = ?";
                statement = conn.prepareStatement(query);
                statement.setInt(1,id);
                statement.execute();
            } catch (SQLException se) {
                se.printStackTrace();
            } finally {
                try {statement.close();} catch (Exception e) {e.printStackTrace();}
                try {conn.close();} catch (Exception e) {e.printStackTrace();}
            }
        }
    }
    
    static private class QuestionRank implements Comparable<QuestionRank> {
        protected Question question;
        protected double rank;
        
        /**
         * Simple class for ordering of Questions
         */
        public QuestionRank(Question question, double rank) {
            this.question = question;
            this.rank = rank;
        }
        
        @Override
        public int compareTo(QuestionRank qr) {
            if (qr.rank == rank) {
                return question.getTimestamp().compareTo(qr.question.getTimestamp());
            }
            return rank > qr.rank ? 1 : -1;
        }

        @Override
        public boolean equals(Object o) {
            if (o instanceof QuestionRank) {
                QuestionRank qr = (QuestionRank) o;
                return qr.rank == rank && question.getTimestamp().equals(qr.question.getTimestamp());
            } else {
                return false;
            }
        }

        @Override
        public int hashCode() {
            int hash = 5;
            hash = 97 * hash + (this.question != null ? this.question.hashCode() : 0);
            hash = 97 * hash + (int) (Double.doubleToLongBits(this.rank) ^ (Double.doubleToLongBits(this.rank) >>> 32));
            return hash;
        }
    }}
