/*
 * BaseServlet.java
 *
 * Created on January 8, 2006, 1:48 PM
 */

package info.heinzelnisse.www;

import info.heinzelnisse.forum.User;
import java.io.*;
import java.util.logging.Logger;

import javax.servlet.*;
import javax.servlet.http.*;

/**
 *
 * @author Heiko Klein
 * @version
 */
public abstract class BaseServlet extends HttpServlet {
    static private Logger logger = Logger.getLogger(BaseServlet.class.getName());
    
    /** Processes requests for both HTTP <code>GET</code> and <code>POST</code> methods.
     * @param request servlet request
     * @param response servlet response
     */
    final protected void processRequest(HttpServletRequest request, HttpServletResponse response)
    throws ServletException, IOException {
        long startTimeMillis = System.currentTimeMillis();
        if (request.getCharacterEncoding() == null) request.setCharacterEncoding("UTF-8");
        ResponseCache responseCache = new ResponseCache();
        
        Cookie[] cookies = request.getCookies();
        if (cookies == null) {
            cookies = new Cookie[0];
        }
        // detect and set the language
        UserSettingsBean userSettings = (UserSettingsBean) request.getSession().getAttribute("userSettings");
        if (getUserSettings(request) == null) {
            userSettings = new UserSettingsBean();
            setUserSettings(request, userSettings);
            for (Cookie cookie : cookies) {
                if (cookie.getName().equals("identity")) {
                    // 0 is displayName, 1 is password
                    String cookieVal = "";
                    javax.mail.internet.MimeUtility.javaCharset("UTF-8");
                    cookieVal = javax.mail.internet.MimeUtility.decodeText(cookie.getValue());
                    String[] values = cookieVal.split("&");
                    if (values.length >= 2) {
                        User user = new User(values[0], values[1]);
                        if (user != null) {
                            userSettings.setUser(user);
                            cookie.setMaxAge(60*60*24*365);
                            cookie.setPath("/");
                        }
                        // invalid user, delete cookie
                        // cookie.setMaxAge(0);
                    } else {
                        // invalid cookie, delete
                        // cookie.setMaxAge(0);
                    }
                    responseCache.addCookie(cookie);
                    logger.fine("setting new timeout for cookie identity");
                }
            }
        }
        // delete the User on logout
        String logout = request.getParameter("logout");
        if (logout != null) {
            for (Cookie cookie : cookies) {
                if (cookie.getName().equals("identity")) {
                    cookie.setMaxAge(0); // 0 deletes cookie
                    cookie.setPath("/");
                    responseCache.addCookie(cookie);
                    logger.fine("deletion of cookie identity");
                    continue;
                }
            }
            userSettings.setUser(new User(0));
        }
        
        String contextPath = request.getContextPath();
        String uri = request.getRequestURI();
        getUserSettings(request).setLangByContextPathAndURI(contextPath,uri);
        
        
        // select the card by getCard (of implementation!)
        CardBean cardb = new CardBean();
        cardb.setCard(getCard());
        request.setAttribute("card", cardb);
        request.setAttribute("title", getTitle(request));
        
        // the header
        responseCache.addIncludePage("/WEB-INF/fragments/header.jsp");
        if (hasLogin()) {
            responseCache.addIncludePage("/WEB-INF/fragments/login.jsp");
        }
        
        // see implementation
        processRequestBody(request, responseCache);
        response.setCharacterEncoding("UTF-8");
        response.setContentType("text/html;charset=UTF-8");
        for (Cookie cookie : responseCache.getCookies()) {
            response.addCookie(cookie);
        }
        if (responseCache.getErrorCode() != HttpServletResponse.SC_OK) {
            response.sendError(responseCache.getErrorCode(), responseCache.getErrorMessage());
        } else if (responseCache.getRedirectPage() == null) {
            for (String jspPage : responseCache.getIncludePages()) {
                logger.fine("dispatching: " + jspPage);
                RequestDispatcher dispatcher = request.getRequestDispatcher(jspPage);
                dispatcher.include(request, response);
            }
            // the footer
            request.setAttribute("requestTime", Long.valueOf(System.currentTimeMillis() - startTimeMillis));
            RequestDispatcher dispatcher = request.getRequestDispatcher("/WEB-INF/fragments/footer.jsp");
            dispatcher.include(request, response);
        } else {
            try {
                response.sendRedirect(responseCache.getRedirectPage());
            } catch (IllegalStateException ise) {
                logger.warning(ise.toString() + " couldn't redirect to "+responseCache.getRedirectPage());
            }
        }
        if (request.getParameter("noSession") != null) {
            request.getSession().invalidate();
        }
    }
    
    
    /**
     * body of the Heinzelpage
     */
//    abstract protected void processRequestBody(HttpServletRequest request, HttpServletResponse response)
//    throws ServletException, IOException;
    
    /**
     * parameters should be added and read from the request
     * the jsp pages to include should be added to the includePages
     * the redirectPage should only be changed (add String), when a redirect is needed
     */
    abstract protected void processRequestBody(HttpServletRequest request, ResponseCache responseCache);
    
    /**
     * defaults to Dict, implementing methods should override this function
     */
    protected CardBean.Card getCard() {
        return CardBean.Card.Dict;
    }
    
    /**
     * title of the current page, depending of the requests language
     */
    protected String getTitle(HttpServletRequest request) {
        UserSettingsBean sb = getUserSettings(request);
        if ("de".equals(sb.getLang()))
            return "Heinzelnisse - Wörterbuch Norwegisch Deutsch";
        else
            return "Heinzelnisse - Ordbok Norsk Tysk";
    }
    
    
    /**
     * defaults to true, implementing classes might remove the 'login' via this page
     */
    protected Boolean hasLogin() {
        return true;
    }
    
    // <editor-fold defaultstate="collapsed" desc="HttpServlet methods. Click on the + sign on the left to edit the code.">
    /** Handles the HTTP <code>GET</code> method.
     * @param request servlet request
     * @param response servlet response
     */
    @Override
    final protected void doGet(HttpServletRequest request, HttpServletResponse response)
    throws ServletException, IOException {
        processRequest(request, response);
    }
    
    /** Handles the HTTP <code>POST</code> method.
     * @param request servlet request
     * @param response servlet response
     */
    @Override
    final protected void doPost(HttpServletRequest request, HttpServletResponse response)
    throws ServletException, IOException {
        processRequest(request, response);
    }
    
    /** Returns a short description of the servlet.
     */
    @Override
    public String getServletInfo() {
        return "Base Servlet for all HeinzelPages";
    }
    // </editor-fold>
    
    public UserSettingsBean getUserSettings(HttpServletRequest request) {
        return (UserSettingsBean) request.getSession().getAttribute("userSettings");
    }
    private void setUserSettings(HttpServletRequest request, UserSettingsBean userSettings) {
        request.getSession().setAttribute("userSettings", userSettings);
    }
}
