/*
 * SearchBean.java
 *
 * Created on October 22, 2006, 11:20 AM
 *
 */
package info.heinzelnisse.www;

import info.heinzelnisse.dictionary.CanooBokmaalLinkDictUtils;
import info.heinzelnisse.dictionary.CanooBokmaalLinkTranslation;
import info.heinzelnisse.dictionary.DictUtils;
import info.heinzelnisse.forum.Answer;
import info.heinzelnisse.forum.Question;
import info.heinzelnisse.wiki.Page;
import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.TreeSet;
import java.util.logging.Logger;

/**
 *
 * @author Heiko Klein
 */
public class SearchBean {
    static private Logger logger = Logger.getLogger(SearchBean.class.getName());
    
    private SearchSettingsBean ssb = new SearchSettingsBean();
    // all cache should be a Collection and the name must start with cache
    // this will be cleared with addRequestParameters or by clearCaches
    List<String> cacheWikiPageNames = null;
    List<Question> cacheForumQuestions = null;
    List<CanooBokmaalLinkTranslation> cacheDeTranslations = null;
    List<CanooBokmaalLinkTranslation> cacheNoTranslations = null;
    List<String> cacheDePhonetics = null;
    List<String> cacheNoPhonetics = null;
    List<String> cacheBokmaalWords = null;
    List<String> cacheNynorskWords = null;
    
    /**
     * Creates a new instance of SearchBean
     */
    public SearchBean() {
    }

    /**
     * result list of wiki search (list of page names)
     */
    public List<String> getWikiPageNames() {
        if (ssb.isWikiSearch()) {
            if (cacheWikiPageNames == null) {
                cacheWikiPageNames = Page.searchPages(ssb.getSearchItem());
            }
            return cacheWikiPageNames;
        } else {
            return Collections.emptyList();
        }
    }

    /**
     * return the (modified) forumQuestions for this search
     */
    public List<Question> getForumQuestions() {
        if (ssb.isForumSearch() || ssb.isForumKeywordSearch()) {
            if (cacheForumQuestions == null) {
                List<Question> questions = Question.searchByKeyword(ssb.getSearchItem()); // forumSearch implies keywordSearch

                if (ssb.isForumSearch()) {
                    questions.addAll(Question.searchByText(ssb.getSearchItem()));
                }
                // removing duplicates
                questions = new ArrayList<Question>(new LinkedHashSet<Question>(questions));
                // show max 30
                if (questions.size() > 30) {
                    questions.subList(30, questions.size()).clear();
                }
                // mark searchWord in red
                for (Question q : questions) {
                    String text = q.getText();
                    text = text.replaceAll("\\Q" + ssb.getSearchItem() + "\\E", "<font color=\"red\">" + ssb.getSearchItem() + "</font>");
                    q.setText(text);
                    for (Answer a : q.getAnswers()) {
                        text = a.getText();
                        text = text.replaceAll("\\Q" + ssb.getSearchItem() + "\\E", "<font color=\"red\">" + ssb.getSearchItem() + "</font>");
                        a.setText(text);
                    }
                }
                cacheForumQuestions = questions;
            }
            return cacheForumQuestions;
        } else {
            return Collections.emptyList();
        }
    }

    /**
     * get translations from german to norwegian
     */
    public List<CanooBokmaalLinkTranslation> getDeTranslations() {
        if (ssb.isDictDeNoSearch()) {
            if (cacheDeTranslations == null) {
                CanooBokmaalLinkDictUtils du = CanooBokmaalLinkDictUtils.getInstance();
                if (ssb.isDictExactSearch()) {
                    cacheDeTranslations = du.dbSearch(ssb.getSearchItem(), DictUtils.DictLang.DE);
                } else {
                    cacheDeTranslations = du.dbSearchStart(ssb.getSearchItem(), DictUtils.DictLang.DE);
                }
            }
            return cacheDeTranslations;
        }
        return Collections.emptyList();
    }

    /**
     * get translations from norwegian to german
     */
    public List<CanooBokmaalLinkTranslation> getNoTranslations() {
        if (ssb.isDictNoDeSearch()) {
            if (cacheNoTranslations == null) {
                CanooBokmaalLinkDictUtils du = CanooBokmaalLinkDictUtils.getInstance();
                if (ssb.isDictExactSearch()) {
                    cacheNoTranslations = du.dbSearch(ssb.getSearchItem(), DictUtils.DictLang.NO);
                } else {
                    cacheNoTranslations = du.dbSearchStart(ssb.getSearchItem(), DictUtils.DictLang.NO);
                }
            }
            return cacheNoTranslations;
        }

        return Collections.emptyList();
    }

    /**
     * get phonetic similar words in german
     */
    public List<String> getDePhonetics() {
        if (ssb.isDictDeNoSearch()) {
            if (cacheDePhonetics == null) {
                if (ssb.isDictPhoneticSearch()) {
                    cacheDePhonetics = CanooBokmaalLinkDictUtils.getInstance().dbSearchSimilarWords(ssb.getSearchItem(), CanooBokmaalLinkDictUtils.DictLang.DE);
                    for (CanooBokmaalLinkTranslation t : getDeTranslations()) {
                        cacheDePhonetics.remove(t.getWord()); // remove on list might be slow, but lists are usually small

                    }
                }
            }
            return cacheDePhonetics;
        }
        return Collections.emptyList();
    }

    /**
     * get phonetic similar words in norwegian
     */
    public List<String> getNoPhonetics() {
        if (ssb.isDictNoDeSearch()) {
            if (cacheNoPhonetics == null) {
                if (ssb.isDictPhoneticSearch()) {
                    cacheNoPhonetics = CanooBokmaalLinkDictUtils.getInstance().dbSearchSimilarWords(ssb.getSearchItem(), CanooBokmaalLinkDictUtils.DictLang.NO);
                    for (CanooBokmaalLinkTranslation t : getNoTranslations()) {
                        cacheNoPhonetics.remove(t.getWord()); // remove on list might be slow, but lists are usually small

                    }
                }
            }
            return cacheNoPhonetics;
        }
        return Collections.emptyList();
    }

    public List<String> getBokmaalWords() {
        if (ssb.isDictBokmaalSearch()) {
            if (cacheBokmaalWords == null) {
                // make sure words in translation don't appear in bokmaal-words, too
                TreeSet<String> bokWords = new TreeSet<String>(CanooBokmaalLinkDictUtils.getInstance().dbSearchSimilarBokmaalWords(ssb.getSearchItem()));
                for (CanooBokmaalLinkTranslation trans : getNoTranslations()) {
                    bokWords.remove(trans.getWord());
                }
                cacheBokmaalWords = new ArrayList<String>(bokWords);
            }
        }
        return cacheBokmaalWords;
    }

    public List<String> getNynorskWords() {
        if (ssb.isDictNynorskSearch()) {
            if (cacheNynorskWords == null) {
                cacheNynorskWords = CanooBokmaalLinkDictUtils.getInstance().dbSearchSimilarNynorskWords(ssb.getSearchItem());
            }
        }
        return cacheNynorskWords;
    }


    /**
     * clear all declared fields starting with cache (must be Collection)
     */
    public void clearCaches() {
        for (Field f : SearchBean.class.getDeclaredFields()) {
            if (f.getName().matches("cache.*")) {
                try {
                    Object obj = f.get(this);
                    if (obj != null) {
                        if (obj instanceof Collection) {
                            f.set(this, null);
                        } else {
                            logger.warning(f.getName() + " not instance of Collection!");
                        }
                    }
                } catch (IllegalArgumentException ex) {
                    ex.printStackTrace();
                } catch (IllegalAccessException ex) {
                    ex.printStackTrace();
                }
            }
        }
    }

    public SearchSettingsBean getSsb() {
        return ssb;
    }

    public void setSsb(SearchSettingsBean ssb) {
        this.ssb = ssb;
    }
    
    
    /// below follow getter from ssb for convenience when working with jsps
    
    public boolean isWikiSearch() {
        return ssb.isWikiSearch();
    }

    public boolean isForumSearch() {
        return ssb.isForumSearch();
    }

    public boolean isForumKeywordSearch() {
        return ssb.isForumKeywordSearch();
    }

    public boolean isDictExactSearch() {
        return ssb.isDictExactSearch();
    }

    public boolean isDictPhoneticSearch() {
        return ssb.isDictPhoneticSearch();
    }

    public boolean isDictBokmaalSearch() {
        return ssb.isDictBokmaalSearch();
    }

    public boolean isDictNynorskSearch() {
        return ssb.isDictNynorskSearch();
    }

    public List<SearchSettingsBean.SearchResultPage> getSearchResultOrder() {
        return ssb.getSearchResultOrder();
    }

    /**
     * @return location of jsp-pages to process the search-results
     */
    public List<String> getSearchResultOrderedJSPs() {
        return ssb.getSearchResultOrderedJSPs();
    }

    public String getSearchItem() {
        return ssb.getSearchItem();
    }

    public boolean isDictDeNoSearch() {
        return ssb.isDictDeNoSearch();
    }

    public boolean isDictNoDeSearch() {
        return ssb.isDictNoDeSearch();
    }

    public boolean isSuggestion() {
        return ssb.isSuggestion();
    }
}
