/*
 * Wiki.java
 *
 * Created on July 16, 2006, 7:23 PM
 */

package info.heinzelnisse.www;

import info.heinzelnisse.forum.User;
import info.heinzelnisse.wiki.Page;
import info.heinzelnisse.wiki.PageVersionBean;
import info.heinzelnisse.www.wiki.ChangedPagesBean;
import info.heinzelnisse.www.wiki.EditBean;
import info.heinzelnisse.www.wiki.PageBean;
import info.heinzelnisse.www.wiki.SearchPagesBean;
import info.heinzelnisse.www.wiki.VersionBean;
import java.io.*;
import java.net.URLDecoder;
import java.net.URLEncoder;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Logger;

import javax.servlet.*;
import javax.servlet.http.*;

/**
 *
 * @author Heiko Klein
 * @version
 */
public class Wiki extends BaseServlet {
    enum Task {ReadPage, EditPage, NewPage, VersionPage, SearchPages, SavePage, RecentChanges};
    private Map<String, Task> tasks = new HashMap<String, Task>();
    private static Logger logger = Logger.getLogger(Wiki.class.getName());
    
    /** Returns a short description of the servlet.
     */
    @Override
    public String getServletInfo() {
        return "Heinzelnisse Wiki";
    }
    
    protected void processRequestBody(HttpServletRequest request, ResponseCache responseCache) {
        SearchSettingsBean ssb = (SearchSettingsBean) request.getSession().getAttribute("wikiSearchSettings");
        if (ssb == null) {
            logger.fine("initialize new wiki-searchbean");
            ssb = new SearchSettingsBean();
            // set defaults for dict
            ssb.setDictDeNoSearch(true);
            ssb.setDictNoDeSearch(true);
            ssb.setDictExactSearch(true);
            ssb.setDictPhoneticSearch(true);
            ssb.setWikiSearch(true);
            ssb.setSuggestion(true);
            ssb.setSearchResultOrder(SearchSettingsBean.SearchResultPage.Wiki, SearchSettingsBean.SearchResultPage.Dict, SearchSettingsBean.SearchResultPage.Forum);
            request.getSession().setAttribute("wikiSearchSettings", ssb);
        }
        ssb.addRequestParameters(request);
        SearchBean sb = new SearchBean();
        sb.setSsb(ssb);
        request.setAttribute("searchSettings", sb);
        logger.fine(".State paramter: " + request.getParameter(".State"));
        Task toDo = tasks.get(request.getParameter(".State"));
        if (toDo == null) toDo = Task.ReadPage;
        logger.fine("Task: "+ toDo);
        switch (toDo) {
            case ReadPage: readPage(request, responseCache); break;
            case EditPage: editPage(request, responseCache); break;
            case SavePage: savePage(request, responseCache); break;
            case SearchPages: break; // automatically done by searchbean, nothing left to do
            case VersionPage: versionPage(request, responseCache); break;
            case RecentChanges: recentChangePage(request, responseCache); break;
            case NewPage: newPage(request, responseCache); break;
            default: readPage(request, responseCache);
        }
    }
    
    @Override
    public void init()
    throws ServletException {
        // edit the tasks-strings
        tasks.put("read", Task.ReadPage);
        tasks.put("edit", Task.EditPage);
        tasks.put("save", Task.SavePage);
        tasks.put("versions", Task.VersionPage);
        tasks.put("new", Task.NewPage);
        tasks.put("recentChanges", Task.RecentChanges);
        tasks.put("search", Task.SearchPages);
    }
    
    
    @Override
    protected CardBean.Card getCard() {
        return CardBean.Card.Wiki;
    }
    
    @Override
    protected String getTitle(HttpServletRequest request) {
        UserSettingsBean sb = getUserSettings(request);
        String pageName = getPageName(request);
        pageName = (pageName == null) ? "" : ": " +pageName;
        if ("de".equals(sb.getLang()))
            return "Heinzelnisse - Sprachthemen"+pageName;
        else
            return "Heinzelnisse - Språktemaer"+pageName;
    }

    
    private void readPage(HttpServletRequest request, ResponseCache responseCache) {
        PageBean wpb = new PageBean();
        String pageName = getPageName(request);
        if ((pageName == null) || pageName.trim().equals("") || pageName.trim().equals("null")) {
            // still lots of requests to "null" (earlier implementation?)
            pageName = "emptyPage";
        }
        logger.fine("readPage: "+ pageName);
        Page page = Page.getPage(pageName);
        if (page == null) {
            try {
                String pageUriX = URLDecoder.decode(request.getRequestURI(), "iso8859-13");
                String pageNameX = pageUriX.replaceAll(".*/","");
                // trying a (wrong) iso8859 request
                if (Page.getPage(pageNameX) != null) {
                    logger.info("redirect to iso8859 pagename: "+URLEncoder.encode(pageNameX, "UTF-8"));
                    responseCache.setRedirectPage(URLEncoder.encode(pageNameX, "UTF-8"));
                    return;
                }
            } catch (UnsupportedEncodingException uee) {
                logger.warning(uee.toString());
            }
            notFoundPage(request, responseCache, pageName);
            return;
        }
        logger.fine("got page: " + page.getName());
        
        // fetch an old version if requested
        String version = request.getParameter("version");
        if ((version != null) && (!version.trim().equals(""))) {
            int iVersion = new Integer(version).intValue();
            if ((iVersion > 0) && (iVersion < page.getVersion())) {
                page = Page.getOldVersion(pageName, iVersion);
                wpb.setOldVersion(true);
            }
        }
        
        wpb.setPage(page);
        request.setAttribute("wikiPage", wpb);
        responseCache.addIncludePage("/WEB-INF/fragments/wiki/read.jsp");
    }
    
    private void notFoundPage(HttpServletRequest request, ResponseCache responseCache, String pageName) {
        logger.warning(pageName + " not found");
        request.setAttribute("pageName", getPageName(request));
        responseCache.addIncludePage("/WEB-INF/fragments/wiki/notFound.jsp");
    }
    
    private void editPage(HttpServletRequest request, ResponseCache responseCache) {
        if (isValidUser(request, responseCache)) {
            String pageName = getPageName(request);
            if (pageName == null) {
                logger.warning("pageName not found for edit"); // should not happen in normal use
                readPage(request, responseCache);
            } else {
                logger.fine("Editing wiki page: " + pageName);
                EditBean web = new EditBean();
                request.getSession().setAttribute("wikiEdit", web);
                web.setPageName(pageName);
                String content = request.getParameter("content");
                if ((content != null) && (!"".equals(content))) {
                    web.setContent(content);
                } else {
                    Page page = Page.getPage(pageName);
                    if (page != null) {
                        web.setContent(page.getContent());
                        if (page.isAdmin() && (!getUserSettings(request).getUser().isAdmin())) {
                            // no adminUser, no edit!
                            request.getSession().removeAttribute("wikiEdit");
                            readPage(request, responseCache);
                            return;
                        }
                    }
                }
                responseCache.addIncludePage("/WEB-INF/fragments/wiki/edit.jsp");
            }
        } else {
            readPage(request, responseCache);
        }
    }
    
    private void savePage(HttpServletRequest request, ResponseCache responseCache) {
        if (isValidUser(request, responseCache)) {
            EditBean web = (EditBean) request.getSession().getAttribute("wikiEdit");
            if (web != null) {
                String content = web.getContent();
                String pageName = web.getPageName();
                if ((content != null) && (pageName != null) && (!"".equals(pageName))) {
                    if (content.matches("^\\s*$")) {
                        logger.fine("content is empty string");
                    } else if (pageName.matches("^\\s*$")) {
                        logger.fine("pageName is empty string");
                    } else {
                        // Page doesn't update same content
                        Page.createOrUpdate(pageName, content, getUserSettings(request).getUser().getId(),new Date(System.currentTimeMillis()), request.getRemoteAddr());
                    }
                }
            }
        }
        request.getSession().removeAttribute("wikiEdit"); // clean-up
        readPage(request, responseCache);
    }
    
    private String getPageName(HttpServletRequest request) {
        String pageName = request.getParameter("pageName");
        if (pageName == null) {
            pageName = request.getPathInfo();
            if (pageName != null) {
                pageName = pageName.replaceFirst("/", ""); // remove leading slash
            } else {
                logger.fine("no pageName");
            }
        }
        return pageName;
    }
    /**
     * test user and add warning to web-page
     */
    private boolean isValidUser(HttpServletRequest request, ResponseCache responseCache) {
        User user = getUserSettings(request).getUser();
        if (user.isValid()) {
            return true;
        } else {
            logger.fine("invalid user");
            responseCache.addIncludePage("/WEB-INF/fragments/wiki/invalidUser.jsp");
            return false;
        }
    }
    
    private void newPage(HttpServletRequest request, ResponseCache responseCache) {
        if (isValidUser(request, responseCache)) {
            responseCache.addIncludePage("/WEB-INF/fragments/wiki/newPage.jsp");
        } else {
            readPage(request, responseCache);
        }
    }
    
    private void versionPage(HttpServletRequest request, ResponseCache responseCache) {
        String pageName = getPageName(request);
        logger.fine("getting version for: " + pageName);
        List<PageVersionBean> versions = Page.getVersions(pageName);
        VersionBean wvb = new VersionBean();
        wvb.setPageName(pageName);
        logger.fine("found " + versions.size() + " versions");
        wvb.setVersions(versions);
        request.setAttribute("wikiVersions", wvb);
        responseCache.addIncludePage("/WEB-INF/fragments/wiki/version.jsp");
    }
    
    private void recentChangePage(HttpServletRequest request, ResponseCache responseCache) {
        Calendar cal = GregorianCalendar.getInstance();
        Date now = cal.getTime();
        cal.set(Calendar.HOUR, 0);
        cal.set(Calendar.MINUTE, 0);
        cal.set(Calendar.SECOND, 0);
        cal.set(Calendar.MILLISECOND, 0);
        Date midnight = cal.getTime();
        cal.add(Calendar.DAY_OF_YEAR, -1);
        Date yesterday = cal.getTime();
        cal.setTime(midnight);
        cal.add(Calendar.WEEK_OF_YEAR, -1);
        Date lastWeek = cal.getTime();
        cal.setTime(midnight);
        cal.set(Calendar.MONTH, -1);
        Date lastMonth = cal.getTime();
        cal.set(Calendar.MONTH, -2);
        Date last3Month = cal.getTime();
        
        ChangedPagesBean cpb = new ChangedPagesBean();
        cpb.setLastDayPages(Page.getChangedPages(yesterday, now));
        cpb.setLastWeekPages(Page.getChangedPages(lastWeek, yesterday));
        cpb.setLastMonthPages(Page.getChangedPages(lastMonth, lastWeek));
        cpb.setLast3MonthPages(Page.getChangedPages(lastMonth, last3Month));
        request.setAttribute("wikiChangedPages", cpb);
        responseCache.addIncludePage("/WEB-INF/fragments/wiki/changedPages.jsp");
    }
    
    private void searchPages(HttpServletRequest request, ResponseCache responseCache) {
        String searchItem = request.getParameter("search");
        if (searchItem == null) {
            readPage(request, responseCache);
            return;
        }
        
        SearchPagesBean spb = new SearchPagesBean();
        List<String> searchPages = Page.searchPages(searchItem);
        logger.fine("searchPages for searchItem " +searchItem +": "+ searchPages.size() + "results");
        spb.setPageNames(searchPages);
        spb.setSearchItem(searchItem);
        request.setAttribute("wikiSearchPages", spb);
        responseCache.addIncludePage("/WEB-INF/fragments/wiki/searchPages.jsp");
    }
}
